#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2020/06/06

import sys
import time

import mss

if sys.platform == 'win32':
    import pyautogui
    import win32gui

from pyxllib.algo.geo import ltrb2polygon


def lookup_mouse_position():
    """ 查看鼠标位置的工具
    """
    from keyboard import read_key

    left_top_point = None
    while True:
        k = read_key()
        if k == 'ctrl':
            # 定位当前鼠标坐标位置
            # 也可以用来定位区域时，先确定区域的左上角点
            left_top_point = pyautogui.position()
            print('坐标：', *left_top_point)
        elif k == 'alt':
            # 定位区域的右下角点，并输出区域
            p = pyautogui.position()
            print('区域(x y w h)：', left_top_point.x, left_top_point.y, p.x - left_top_point.x, p.y - left_top_point.y)
        elif k == 'esc':
            break
        # keyboard的监控太快了，需要暂停一下
        time.sleep(0.4)


def type_text(text):
    """ 打印出文本内容

    相比pyautogui.write，这里支持中文等unicode格式

    这种需求一般也可以用剪切板实现，但是剪切板不够独立轻量，可能会有意想不到的副作用
    """
    from pynput.keyboard import Controller

    keyboard = Controller()
    keyboard.type(text)


def clipboard_decorator(rtype='text', *, copy=True, paste=False, typing=False):
    """ 装饰器，能方便地将函数的返回值转移到剪切板文本、富文本，甚至typing打印出来。

    Args:
        rtype: 函数返回值类型
        paste: 是否执行剪切板的粘贴功能
        typing: 是否键入文本内容

    Returns:

    """
    from bs4 import BeautifulSoup
    import pyautogui
    import klembord

    def decorator(func):
        def wrapper(*args, **kwargs):
            # 1 运行函数获得结果
            s = func(*args, **kwargs)

            # 2 复制到剪切板
            if copy:
                if rtype == 'text' and isinstance(s, str):
                    klembord.set_text(s)
                elif rtype == 'html' and isinstance(s, str):
                    s0 = BeautifulSoup(s, 'lxml').text
                    klembord.set_with_rich_text(s0, s)
                elif rtype == 'dict' and isinstance(s, dict):
                    klembord.set(s)

            # 3 输出
            if paste:
                pyautogui.hotkey('ctrl', 'v')  # 目前来看就这个方法最靠谱
            if typing:
                type_text(s)

            return s

        return wrapper

    return decorator


def get_clipboard_content(rich=False, *, head=False):
    """ klembord的get_with_rich_text获取富文本有点小问题，所以自己重写了一个版本
    好在有底层接口，所以也不难改

    :param rich: 是否返回html格式的富文本，默认True
    :param head: 富文本时，是否要返回<html><body>的标记
    """
    import re
    import html

    import klembord
    from klembord import W_HTML, W_UNICODE

    if not rich:
        return klembord.get_text()

    content = klembord.get((W_HTML, W_UNICODE))
    html_content = content[W_HTML]
    if html_content:
        html_content = html_content.decode('utf8')
        html_content = re.search(r'<body>\s*(?:<!--StartFragment-->)?(.+?)(?:<!--EndFragment-->)?\s*</body>',
                                 html_content).group(1)
    else:
        # 没有html的时候，可以封装成html格式返回
        text = content[W_UNICODE].decode('utf16')
        html_content = html.escape(text)

    if head:
        html_content = '<html><body>' + html_content + '</body></html>'

    return html_content


def set_clipboard_content(content, *, rich=False):
    """ 对klembord剪切板的功能做略微的简化 """
    from bs4 import BeautifulSoup
    import klembord

    if rich:
        klembord.set_with_rich_text(BeautifulSoup(content, 'lxml').text, content)
    else:
        klembord.set_text(content)


def grab_pixel_color(radius=0):
    """
    :param radius: 以中心像素为例，计算矩阵范围内的颜色分布情况（未实装）
    """
    from pyxllib.cv.rgbfmt import RgbFormatter
    last_pos = None
    print('精确全体/精确中文')

    while True:
        pos = pyautogui.position()
        if pos == last_pos:
            continue
        else:
            last_pos = pos

        color = RgbFormatter(*pyautogui.pixel(*pos))
        desc = color.relative_color_desc(color_range=2, precise_mode=True)
        print('\r' + f'{color.to_hex()} {desc}', end='', flush=True)
        time.sleep(0.1)


def grab_monitor(order=0, to_cv2=True):
    """ 截屏

    :param int order:
        默认0，截取全部屏幕
        1，截取第1个屏幕
        2，截取第2个屏幕
        ...
    :param to_cv2:
        True, 转成cv2格式的图片
        False, 使用pil格式的图片
    """
    from PIL import Image
    from pyxllib.xlcv import PilImg

    with mss.mss() as sct:
        assert order < len(sct.monitors), '屏幕下标越界'
        monitor = sct.monitors[order]
        sct_img = sct.grab(monitor)
        img = Image.frombytes("RGB", sct_img.size, sct_img.bgra, "raw", "BGRX")
        if to_cv2:
            img = PilImg.read(img).to_cv2_image()  # 返回是np矩阵
        return img


def list_windows(mode=1):
    """ 列出所有窗口

    :param mode:
        0, 列出所有窗口
        1, 只列出有名称，有宽、高的窗口
        str, 指定名称的窗口
            注意，当结果有多个的时候，一般后面那个才是紧贴边缘的窗口位置

    :return: [[窗口名, xywh], [窗口2, xywh]...]
    """
    import win32gui

    ls = []

    def callback(hwnd, extra):
        l, t, r, b = win32gui.GetWindowRect(hwnd)
        w, h = r - l, b - t
        name = win32gui.GetWindowText(hwnd)
        if mode == 0:
            ls.append([name, [l, t, w, h]])
        elif mode == 1 and name and w and h:
            ls.append([name, [l, t, w, h]])
        elif isinstance(mode, str) and name == mode:
            ls.append([name, [l, t, w, h]])

    win32gui.EnumWindows(callback, None)
    return ls


class UiTracePath:
    """ 在window窗口上用鼠标勾画路径 """

    @classmethod
    def from_polygon(cls, points, duration_per_circle=3, num_circles=1):
        """
        让鼠标顺时针围绕任意多边形移动。

        :param points: 多边形顶点的坐标列表，按顺时针或逆时针顺序排列
        :param duration_per_circle: 每圈耗时
        :param num_circles: 总圈数
        """
        num_points = len(points)
        segment_duration = duration_per_circle / num_points  # 每条边的耗时

        for _ in range(num_circles):
            # 顺时针移动鼠标
            for i in range(num_points):
                start = points[i]
                pyautogui.moveTo(start[0], start[1])
                end = points[(i + 1) % num_points]
                # 使用 moveTo 的 duration 参数控制每条边的移动时间
                pyautogui.moveTo(end[0], end[1], duration=segment_duration)

    @classmethod
    def from_ltrb(cls, ltrb, **kwargs):
        points = ltrb2polygon(ltrb)
        return cls.from_polygon(points, **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/02/25 17:54

import cv2
from moviepy.editor import VideoFileClip
from moviepy.editor import cvsecs
import numpy as np
from tqdm import tqdm

from pyxllib.prog.pupil import inject_members
from pyxllib.file.specialist import XlPath
from pyxllib.cv.imhash import get_init_hash, phash
from pyxllib.cv.xlcvlib import xlcv


class XlVideoFileClip(VideoFileClip):

    def get_frame2(self, time_point, *, scale=None):
        """ 官方获得的图片通道是RGB，但是cv2处理的统一规则是BGR，要转换过来

        :param scale: 获取图片后是否要按统一的比例再缩放一下
        """
        frame = self.get_frame(time_point)
        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

        if scale:
            frame = cv2.resize(frame, (0, 0), fx=scale, fy=scale)

        return frame

    def get_frames(self, time_points=None, *, interval_second=0.1, cur_hash=None, head_frame=None, scale=None,
                   filter_mode=2, print_mode=True):
        """ 同时获得多帧图片

        :param time_points: 类list对象，元素是时间点（可以是字符串格式，也可以是数值秒数）
            常见生成方法参考：np.array(0, clip.duration, 0.1)
        :param filter_mode: 过滤强度
            因为times如果取的非常密集，比如每0.1秒取一帧，会有很大概率是重复、相近的图
            此时可以用这个参数控制误差在多大以内的图会保留
            使用phash判断帧之间相似度，只要误差在该设定阈值内，都会被跳过
            可以设置0，不过这样过滤程度是最低的
            也可以输入-1，表示完全不过滤，不计算phash。当这种情况也不太必要用这个函数了，可以直接普通循环。
            也可以故意设置的特别大（phash最大差距是64），那这样相当于差异很大的也过滤掉了
        :param head_frame: 提供一张初始图片，主要是配合filter_mode用于去重的
        :param cur_hash: 类似head_frame作用，但直接提供hash值
        :return: 使用yield机制，防止图片数据太多，内存爆炸了
            当filter_mode>=0时，返回 时间点time_point和对应的图片im
        """
        if cur_hash is None:
            if head_frame is None:
                cur_hash = get_init_hash()
            else:
                cur_hash = phash(head_frame)

        if time_points is None:
            time_points = np.arange(0, self.duration, interval_second)

        for time_point in tqdm(time_points, disable=not print_mode):
            im = self.get_frame2(time_point, scale=scale)

            if filter_mode >= 0:
                last_hash, cur_hash = cur_hash, phash(im)
                if cur_hash - last_hash <= filter_mode:
                    continue
                yield time_point, im
            else:
                yield im

    def save_frames(self, out_dir, time_points=None, interval_second=0.1, **kwargs):
        """ 跟get_frames差不多，多了一步自动存储图片文件到目录里

        这个功能是比较定制化的，不是那么泛用，但会有很多相关类似的操作需求，可以参考这里的代码实现
        """
        # 1 time_points
        if time_points is None:
            time_points = np.arange(0, self.duration, interval_second)
        out_dir = XlPath(out_dir)

        # 2 按时间点取图片，保存文件
        for time_point, frame in self.get_frames(time_points, **kwargs):
            tt = int(time_point * 10)
            m, s, ms = tt // 600, (tt // 10) % 60, tt % 10
            xlcv.write(frame, out_dir / f'{m:03}_{s:02}.{ms}.jpg')

    def join_subtitles_image(self, time_points, ltrb_pos=None, *,
                             crop_first_frame=False,
                             filter_mode=2):
        """ 生成字幕拼图

        :param time_points: 在哪些时间点截图
        :param ltrb_pos: 字幕所在位置，没有输入则默认全部全图拼接，
            一般指定上下就行了，即 [None, 640, None, 700]
        :param crop_first_frame:
            False，保留第一帧的完整性
            True，第一帧也只裁剪字幕部分
        :param filter_mode: 对于给出的时间点图片，去除相邻相同的图片
        :return:

        参考用法：
            from pyxllib.file.movielib import VideoFileClip
            clip = VideoFileClip(str(Paths.videos / '觉观01.mp4'))  # 必须要转str类型
            im = clip.join_subtitles_image(np.arange(30, 40, 0.1), [None, 640, None, 695])
            xlcv.write(im, Paths.videos / 'test.jpg')
        """
        # 1 如果有左右裁剪要提前处理
        x1, y1, x2, y2 = ltrb_pos
        if x1 or x2:
            clip = self.crop(x1=x1, x2=x2)
        else:
            clip = self

        # 2 第1帧是否要保留全图
        frame_list = []
        if crop_first_frame:
            head_frame = None
        else:
            time_points = tuple(time_points)
            head_frame = clip.get_frame2(time_points[0])
            frame_list.append(head_frame[:y2])
            head_frame = head_frame[y1:y2]
            time_points = time_points[1:]

        # 3 裁剪字幕区域
        clip = clip.crop(y1=y1, y2=y2)
        for time_point, frame in clip.get_frames(time_points, filter_mode=filter_mode, head_frame=head_frame):
            frame_list.append(frame)

        # 4 拼接完整图
        im = xlcv.concat(frame_list, pad=0)
        return im


inject_members(XlVideoFileClip, VideoFileClip)


def clip_video(input_file, output_file, start_time, end_time):
    start_time = cvsecs(start_time)
    end_time = cvsecs(end_time)
    clip = VideoFileClip(input_file).subclip(start_time, end_time)
    clip.write_videofile(output_file)
    clip.reader.close()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/08/09

import os
import json
from json import dumps as jsonDumps
from json import loads as jsonLoads
from pathlib import Path
from typing import List, Literal
import base64
import logging
from unittest.mock import MagicMock
import math
from types import SimpleNamespace
from typing import Union

import requests
from loguru import logger as loguru_logger
from tqdm import tqdm
from joblib import Parallel, delayed
from openai import OpenAI, AsyncOpenAI

from pyxllib.text.jinjalib import set_template
from pyxllib.file.specialist import download_file


def __1_Chat类的基础功能():
    pass


class AKPool:
    """ 轮询获取api_key """

    def __init__(self, apikeys: list):
        self._pool = self._POOL(apikeys)

    def fetch_key(self):
        return next(self._pool)

    @classmethod
    def _POOL(cls, apikeys: list):
        while True:
            for x in apikeys:
                yield x


class MsgBase:
    role_name: str
    text: str

    def __init__(self, text: str):
        self.text = text

    def __str__(self):
        return self.text

    def __iter__(self):
        yield "role", self.role_name
        yield "content", self.text


system_msg = type("system_msg", (MsgBase,), {"role_name": "system"})
user_msg = type("user_msg", (MsgBase,), {"role_name": "user"})
assistant_msg = type("assistant_msg", (MsgBase,), {"role_name": "assistant"})


class Temque:
    """ 一个先进先出, 可设置最大容量, 可固定元素的队列 """

    def __init__(self, maxlen: int = None):
        self.core: List[dict] = []
        self.maxlen = maxlen or float("inf")

    def _trim(self):
        core = self.core
        if len(core) > self.maxlen:
            dc = len(core) - self.maxlen
            indexes = []
            for i, x in enumerate(core):
                if not x["pin"]:
                    indexes.append(i)
                if len(indexes) == dc:
                    break
            for i in indexes[::-1]:
                core.pop(i)

    def add_many(self, *objs):
        for x in objs:
            self.core.append({"obj": x, "pin": False})
        self._trim()

    def __iter__(self):
        for x in self.core:
            yield x["obj"]

    def pin(self, *indexes):
        for i in indexes:
            self.core[i]["pin"] = True

    def unpin(self, *indexes):
        for i in indexes:
            self.core[i]["pin"] = False

    def copy(self):
        que = self.__class__(maxlen=self.maxlen)
        que.core = self.core.copy()
        return que

    def deepcopy(self):
        ...  # 创建这个方法是为了以代码提示的方式提醒用户: copy 方法是浅拷贝

    def __add__(self, obj: 'list|Temque'):
        que = self.copy()
        if isinstance(obj, self.__class__):
            que.core += obj.core
            que._trim()
        else:
            que.add_many(*obj)
        return que


class Chat:
    """
    [文档](https://lcctoor.github.io/arts/arts/openai2)

    获取api_key:
    * [获取链接1](https://platform.openai.com/account/api-keys)
    * [获取链接2](https://www.baidu.com/s?wd=%E8%8E%B7%E5%8F%96%20openai%20api_key)
    """

    recently_request_data: dict  # 最近一次请求所用的参数
    default_base_url = os.getenv('OPENAI_BASE_URL')
    default_api_key = os.getenv('OPENAI_API_KEY')

    def __init__(self,
                 # kwargs
                 api_key: Union[str, AKPool] = None,
                 base_url: str = None,  # base_url 参数用于修改基础URL
                 timeout=None,
                 max_retries=None,
                 http_client=None,
                 # request_kwargs
                 model: Literal["gpt-4-1106-preview", "gpt-4-vision-preview",
                 "gpt-4", "gpt-4-0314", "gpt-4-0613",
                 "gpt-4-32k", "gpt-4-32k-0314", "gpt-4-32k-0613", "gpt-3.5-turbo",
                 "gpt-4o", "gpt-4-turbo"] = "gpt-4o-mini",
                 # Chat
                 msg_max_count: int = None,
                 # kwargs
                 **kwargs,
                 ):
        """

        :param kwargs: 主要是供OpenAI初始化时使用的参数

        """
        api_key = api_key or self.default_api_key or os.getenv("OPENAI_API_KEY")

        api_base = kwargs.pop('api_base', None)
        base_url = base_url or api_base or self.default_base_url or os.getenv("OPENAI_BASE_URL")
        MsgMaxCount = kwargs.pop('MsgMaxCount', None)
        msg_max_count = msg_max_count or MsgMaxCount

        if base_url: kwargs["base_url"] = base_url
        if timeout: kwargs["timeout"] = timeout
        if max_retries: kwargs["max_retries"] = max_retries
        if http_client: kwargs["http_client"] = http_client

        self.reset_api_key(api_key)
        self._kwargs = kwargs
        self._request_kwargs = {'model': model}
        self._messages = Temque(maxlen=msg_max_count)

    def __1_原作者提供功能(self):
        pass

    def reset_api_key(self, api_key: Union[str, AKPool]):
        if isinstance(api_key, AKPool):
            self._akpool = api_key
        else:
            self._akpool = AKPool([api_key])

    def request(self, text: str = None, **kwargs):
        """
        :param kwargs: 支持本次请求的参数覆盖全局 self._request_kwargs 的参数
        """
        messages = [{"role": "user", "content": text}]
        messages += (kwargs.pop('messages', None) or [])  # 兼容官方包[openai]用户, 使其代码可以无缝切换到[openai2]
        assert messages
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        completion = OpenAI(api_key=api_key, **self._kwargs).chat.completions.create(**{
            **self._request_kwargs,  # 全局参数
            **kwargs,  # 单次请求的参数覆盖全局参数
            "messages": list(self._messages + messages),
            "stream": False,
        })
        answer: str = completion.choices[0].message.content
        self._messages.add_many(*messages, {"role": "assistant", "content": answer})
        return answer

    def stream_request(self, text: str = None, **kwargs):
        messages = [{"role": "user", "content": text}]
        messages += (kwargs.pop('messages', None) or [])  # 兼容官方包[openai]用户, 使其代码可以无缝切换到[openai2]
        assert messages
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        completion = OpenAI(api_key=api_key, **self._kwargs).chat.completions.create(**{
            **self._request_kwargs,  # 全局参数
            **kwargs,  # 单次请求的参数覆盖全局参数
            "messages": list(self._messages + messages),
            "stream": True,
        })
        answer: str = ""
        for chunk in completion:
            if chunk.choices and (content := chunk.choices[0].delta.content):
                answer += content
                yield content
        self._messages.add_many(*messages, {"role": "assistant", "content": answer})

    async def async_request(self, text: str = None, **kwargs):
        messages = [{"role": "user", "content": text}]
        messages += (kwargs.pop('messages', None) or [])  # 兼容官方包[openai]用户, 使其代码可以无缝切换到[openai2]
        assert messages
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        completion = await AsyncOpenAI(api_key=api_key, **self._kwargs).chat.completions.create(**{
            **self._request_kwargs,  # 全局参数
            **kwargs,  # 单次请求的参数覆盖全局参数
            "messages": list(self._messages + messages),
            "stream": False,
        })
        answer: str = completion.choices[0].message.content
        self._messages.add_many(*messages, {"role": "assistant", "content": answer})
        return answer

    async def async_stream_request(self, text: str = None, **kwargs):
        messages = [{"role": "user", "content": text}]
        messages += (kwargs.pop('messages', None) or [])  # 兼容官方包[openai]用户, 使其代码可以无缝切换到[openai2]
        assert messages
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        completion = await AsyncOpenAI(api_key=api_key, **self._kwargs).chat.completions.create(**{
            **self._request_kwargs,  # 全局参数
            **kwargs,  # 单次请求的参数覆盖全局参数
            "messages": list(self._messages + messages),
            "stream": True,
        })
        answer: str = ""
        async for chunk in completion:
            if chunk.choices and (content := chunk.choices[0].delta.content):
                answer += content
                yield content
        self._messages.add_many(*messages, {"role": "assistant", "content": answer})

    def rollback(self, n=1):
        '''
        回滚对话
        '''
        self._messages.core[-2 * n:] = []
        for x in self._messages.core[-2:]:
            x = x["obj"]
            print(f"[{x['role']}]:{x['content']}")

    def pin_messages(self, *indexes):
        '''
        锁定历史消息
        '''
        self._messages.pin(*indexes)

    def unpin_messages(self, *indexes):
        '''
        解锁历史消息
        '''
        self._messages.unpin(*indexes)

    def fetch_messages(self):
        return list(self._messages)

    def add_dialogs(self, *ms: Union[dict, system_msg, user_msg, assistant_msg]):
        '''
        添加历史对话
        '''
        messages = [dict(x) for x in ms]
        self._messages.add_many(*messages)

    def __getattr__(self, name):
        # match name:  # 兼容旧代码
        #     case 'asy_request':
        #         return self.async_request
        #     case 'forge':
        #         return self.add_dialogs
        #     case 'pin':
        #         return self.pin_messages
        #     case 'unpin':
        #         return self.unpin_messages
        #     case 'dump':
        #         return self._dump
        #     case 'load':
        #         return self._load
        # raise AttributeError(name)

        if name == 'asy_request':
            return self.async_request
        elif name == 'forge':
            return self.add_dialogs
        elif name == 'pin':
            return self.pin_messages
        elif name == 'unpin':
            return self.unpin_messages
        elif name == 'dump':
            return self._dump
        elif name == 'load':
            return self._load
        else:
            raise AttributeError(name)

    def _dump(self, fpath: str):
        """ 存档 """
        jt = jsonDumps(self.fetch_messages(), ensure_ascii=False)
        Path(fpath).write_text(jt, encoding="utf8")
        return True

    def _load(self, fpath: str):
        """ 载入存档 """
        jt = Path(fpath).read_text(encoding="utf8")
        self._messages.add_many(*jsonLoads(jt))
        return True

    def __2_扩展功能(self):
        pass

    def add_message(self, message):
        """ 添加单条历史消息记录 """
        self._messages.add_many(message)

    def add_messages(self, messages: list[dict]):
        """ 添加多条历史消息记录 """
        self._messages.add_many(*messages)

    def parse_cont_to_conts(self, in_cont) -> list[dict]:
        """ 自定义的一套内容设置格式，用来简化openai源生的配置逻辑

        :return: 注意虽然看似输入是一个cont，实际可能解析出多个conts，所以返回的是list类型
        """

        def encode_image(image_path):
            """ 读取本地图片文件 """
            with open(image_path, "rb") as image_file:
                return base64.b64encode(image_file.read()).decode('utf-8')

        def parse_image_file(image_file):
            suffix = Path(image_file).suffix[1:].lower()
            if suffix == 'jpg':
                suffix = 'jpeg'
            base64_image = encode_image(image_file)
            image_url = {'url': f"data:image/{suffix};base64,{base64_image}"}
            return image_url

        if isinstance(in_cont, str):
            return [{"type": "text", "text": in_cont}]

        if isinstance(in_cont, dict):
            if 'image_url' in in_cont and isinstance(in_cont['image_url'], str):
                url = in_cont['image_url']
                return [{"type": "image_url", "image_url": {'url': url}}]
            if 'image_file' in in_cont:
                image_url = parse_image_file(in_cont['image_file'])
                in_cont.pop('image_file')
                if in_cont:
                    image_url.update(in_cont)
                return [{"type": "image_url", "image_url": image_url}]

            if 'image_urls' in in_cont:
                return [{"type": "image_url", "image_url": {'url': x}} for x in in_cont['image_urls']]
            if 'image_files' in in_cont:
                return [{"type": "image_url", "image_url": parse_image_file(x)} for x in in_cont['image_files']]

            return [in_cont]
        else:
            raise ValueError

    def parse_conts_to_message(self, conts, role='user'):
        """ 可以用一种简化的格式，来配置用户要提问的信息内容

        :param conts:
            str, 单条文本提问
            list,
                类似这样的数据，可以完全兼容openai的格式
                    [
                    {"type": "text", "text": "What’s in this image?"},
                    {
                      "type": "image_url",
                      "image_url": {
                        "url": "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg",
                      },
                    },
                    ]
              也可以简化：用我这里自定义的简化格式
                'text'    文本格式直接输入
                {'image_url': 'https://...'}
                {'image_file': 'path/to/file'}
        """
        message = {"role": role}

        if isinstance(conts, str):
            message["content"] = conts
        else:
            conts = conts if isinstance(conts, (list, tuple)) else [conts]
            message["content"] = []
            for x in conts:
                message["content"] += self.parse_cont_to_conts(x)

        return message

    def add_message_from_conts(self, conts, role='user'):
        """ 添加content版本的message """
        message = self.parse_conts_to_message(conts, role=role)
        self._messages.add_many(message)

    def add_system_prompt(self, conts):
        self.add_message_from_conts(conts, 'system')

    def query(self, conts=None, response_json=None, **kwargs):
        """ 提交一个请求，并获得回复结果

        :param conts:
            None, content可以不输入，这种一般是外部已经提前配置好各种add_message，这里只要直接请求即可
            str, 常规的文本提问使用
            list, 一般是搭配图片的提问使用
        :param response_json: 是否返回json格式
            注意就算这个参数为True，提示里也要显示说明需要json格式。官方示例是在初始的role=system中配置
        """
        # 1 total_messages
        if conts:
            messages = [self.parse_conts_to_message(conts)]
            messages += (kwargs.pop('messages', None) or [])
        else:
            messages = []
        total_messages = list(self._messages + messages)
        assert total_messages

        if response_json:
            kwargs['response_format'] = {"type": "json_object"}

        # 2 请求结果
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        completion = OpenAI(api_key=api_key, **self._kwargs).chat.completions.create(**{
            **self._request_kwargs,  # 全局参数
            **kwargs,  # 单次请求的参数覆盖全局参数
            "messages": total_messages,
            "stream": False,
        })
        answer: str = completion.choices[0].message.content
        self._messages.add_many(*messages, {"role": "assistant", "content": answer})

        if response_json:
            answer = json.loads(answer)

        return answer

    def query_image(self, prompt,
                    size='1024x1024', model='dall-e-3', quality='standard', n=1,
                    save_file=None,
                    **kwargs):
        """ 生成图片，注意，此功能不支持引用旧历史记录，而是一个独立的记录来请求model获得一张图

        :param save_file: 指定保存到本地某个位置
        """
        self.recently_request_data = {
            'api_key': (api_key := self._akpool.fetch_key()),
        }
        client = OpenAI(api_key=api_key, **self._kwargs)
        response = client.images.generate(
            model=model,
            prompt=prompt,
            size=size,
            quality=quality,
            n=n,
            **kwargs
        )
        image_url = response.data[0].url
        self.add_message_from_conts(prompt, 'user')
        self.add_message_from_conts([{'image_url': image_url}], 'assistant')

        if save_file:
            download_file(image_url, save_file)

        return image_url


class Chat2:
    default_api_key = None
    default_base_url = 'https://beta.gpt4api.plus'

    def __init__(self, api_key=None,
                 gizmo_id=None,
                 *,
                 base_url=None,
                 model='gpt-4o-mini'):
        """
        :param gizmo_id: gpts的id，比如 https://chatgpt.com/g/g-ABCD1234-ce-shi。则id就是"g-ABCD1234"
            后面的gpts的名称变了没关系，前面前缀的id不要变就行
        """
        self.api_key = api_key or self.default_api_key
        self.base_url = base_url or self.default_base_url
        self.gizmo_id = gizmo_id  # 是否有action模块需要点击确认

        self.headers = {
            'Authorization': f'Bearer {self.api_key}'
        }
        self.session_state = {
            'stream': False,
            'model': model,
        }

        self.messages = list()  # 历史记录

    def _upload_file(self, file_path, file_type='my_files', parent_payload=None):
        """ 上传文件

        :param payload: 如果传入外部主payload自动，会自动添加已上传的文件的状态内容信息
        """
        # 1 请求
        url = f"{self.base_url}/chat/uploaded"
        payload = {'type': file_type}
        if 'conversation_id' in self.session_state:
            payload['conversation_id'] = self.session_state['conversation_id']
        files = [
            ('files', (Path(file_path).name, open(file_path, 'rb')))
        ]
        resp = requests.post(url, headers=self.headers, data=payload, files=files)
        resp_json = resp.json()

        # 2 保存状态
        self.session_state['conversation_id'] = resp_json['conversation_id']
        self.messages.append(resp_json)
        if parent_payload:
            key_name = 'attachments' if file_type == 'my_files' else 'parts'
            if key_name not in parent_payload:
                parent_payload[key_name] = []
            parent_payload[key_name].append(resp_json[key_name[:-1]])

        return resp_json

    def _upload_image(self, image_path, parent_payload=None):
        """ 上传图片 """
        return self._upload_file(image_path, 'multimodal', parent_payload=parent_payload)

    def query(self, message, *, files=None, images=None):
        # 1 预备
        if self.gizmo_id:
            url = f"{self.base_url}/chat/action"
        else:
            url = f"{self.base_url}/chat/all-tools"

        headers = self.headers.copy()
        # headers['Content-Type'] = 'application/json'

        # 2 api初步配置
        payload = self.session_state.copy()
        payload['message'] = message
        if self.gizmo_id:
            payload['gizmo_id'] = self.gizmo_id

        # 3 添加文件
        if files:
            if isinstance(files, str):
                files = [files]
            for file in files:
                self._upload_file(file, parent_payload=payload)

        # 4 添加图片
        if images:
            if isinstance(images, str):
                images = [images]
            for image in images:
                self._upload_image(image, parent_payload=payload)

        # 5 请求
        # resp = requests.post(url, headers=headers, data=json.dumps(payload))
        resp = requests.post(url, headers=headers, json=payload)  # 我觉得应该可以改这样调用，但还没测试
        resp_json = resp.json()

        # 6 保存状态
        self.session_state['conversation_id'] = resp_json['conversation_id']
        self.session_state['parent_message_id'] = resp_json['message_id']
        self.messages.append(resp_json)

        return '\n\n'.join(resp_json['contents'][-1]['message']['content']['parts'])


class DifyChat:
    default_base_url = 'http://localhost'

    def __init__(self, app_key, *,
                 conversation_id=None,
                 user=None,
                 base_url=None):
        self.app_key = app_key
        self.base_url = base_url or self.default_base_url

        self.headers = {
            'Authorization': f'Bearer {self.app_key}',
        }

        self.conversation_state = {
            'inputs': {},
            # 'response_mode': 'streaming',
            'conversation_id': conversation_id or '',
            'user': user or 'code4101',
        }

    def query(self, query, *, files=None, images=None):
        """ files、images等后期再加
        """
        # 1 请求
        payload = self.conversation_state.copy()
        payload['query'] = query
        resp = requests.post(f'{self.base_url}/v1/chat-messages',
                             headers=self.headers, json=payload)
        resp_json = resp.json()  # 如果需要可以打印这个字典看结构

        # 2 保存状态
        self.conversation_state['conversation_id'] = resp_json.get('conversation_id', '')

        return resp_json['answer']


def __2_信息摘要():
    pass


# 用来记录一些常用的提示词
xlprompts = SimpleNamespace()

# 通用的信息压缩
xlprompts.compress_content_common = """
角色: 你是一个专业的信息摘要助手，擅长快速提取和总结大量文本中的关键信息。
任务: 你的任务是对输入的"从全文中节选的部分文本内容"，提取出最重要的要点，确保摘要简洁明了，保持原意并覆盖主要内容。你需要注意以下几点：

重点提取: 确保抓住文本中的关键主题、观点和结论。
简明扼要: 摘要应尽量简短，但不能遗漏重要信息。
客观性: 保持中立，不加入任何个人意见或评论。
上下文理解: 理解并保留信息的背景和细节，以免丢失关键上下文。
适应性: 无论输入文本的类型、领域或长度如何，你都能有效生成高质量的摘要。
""".strip()

# 根据用户的问题，进行信息压缩
xlprompts.compress_content_with_query = set_template("""
角色: 你是一个智能的信息摘要助手，能够根据用户提供的查询<query>灵活调整摘要的侧重点。
任务: 你的任务是对输入的"从全文中节选的部分文本内容"中，提取出与用户查询<query>最相关的内容，并根据这些重点信息生成一个简洁的摘要。
确保摘要既包含用户感兴趣的内容，又能保留必要的背景信息。你需要注意以下几点：

查询相关性: 优先提取与用户查询相关的内容，将其放在摘要的核心部分。
简洁性: 保证摘要清晰简洁，避免冗长和不必要的细节。
全面性: 在满足查询相关性的前提下，尽量保留与主题相关的背景信息。
上下文理解: 理解用户查询的意图，并在生成摘要时结合上下文做出适当的调整。
适应性: 无论输入文本的类型、领域或长度如何，你都能有效生成侧重用户查询的高质量摘要。

<query>
{{ query }}
""")

xlprompts.compress_qq_message = """
### 目的
你是一个信息整理助手，这里任务是负责整理我的QQ聊天记录，便于后续查阅和使用，特制定以下规则。

### 整理步骤
1. **按时间线排序**：将聊天记录按时间顺序整理。
2. **聚合信息**：将内容相似或相关的聊天记录聚合，避免过于细碎。
3. **引入人员信息**：在描述中适当引入发言人员的名字，确保信息来源清晰。
4. **内容简洁**：确保每条记录简洁明了，保留关键内容。

### 具体格式
1. **时间格式**：以“YYYY-MM-DD HH:MM:SS”格式记录时间。
2. **内容聚合**：
   - **主题聚合**：将同一主题的聊天记录合并整理。
   - **人名引入**：在描述中加入发言人员名字，例如“韩锦锦提到…”，“代号4101解释…”
3. **示例**：
```markdown
### 2024-02-21
- **童浩关于空表的处理**：童浩提到已经将空表都补上，方便自动测试，@23研一 童浩 数据已修改对应实际内容，避免遇到没有表格的题目。
- **关于条件格式**：代号4101询问是否有办法做出多条件的约束，22软工-郑创鸿表示条件格式不太好做，23研一 童浩确认空表已补上。代号4101提到WPS偏爱条件格式，建议确认完成后查看结果。
- **模型数据准备**：代号4101建议使用GPT-4帮助生成正确数据，并呼吁团队注意在一些题目中替换正确数据以进行测试。

### 2024-02-22
- **多轮回答格式调整**：代号4101通知韩锦锦关于多轮回答的格式调整已完成，可以开始标注。
```

### 注意事项
- **避免冗长**：每条记录保持简洁，去除无关信息。
- **保留原意**：整理过程中，确保保留原始信息的准确性和完整性。
""".strip()


class TextSpliter:
    """ 一组文本切分算法 """

    @classmethod
    def avg(cls, content, segment_len, overlap_len):
        """ 均匀切分，不过这个具体实现还是比较基础版的

        >> TextSpliter.avg('123456789', 10, 0)
        ['123456789']
        >> TextSpliter.avg('123456789', 9, 0)
        ['12345', '6789']
        >> TextSpliter.avg('123456789', 9, 1)
        ['12345', '56789']
        >> TextSpliter.avg('123456789', 9, 2)
        ['12345', '45678', '789']
        >> TextSpliter.avg('123456789', 5, 3)
        ['12345', '34567', '56789']
        """
        content_len = len(content)
        num = content_len // segment_len + 1  # 目标切分数量
        segment_len = math.ceil(content_len / num)  # 重置后的segment_len
        return [content[i:i + segment_len] for i in range(0, content_len - overlap_len, segment_len - overlap_len)]


class CompressContent:

    @classmethod
    def basic(cls, contents, extra_prompts,
              segment_len=5000, overlap_len=200, *,
              max_workers=1,
              expect_len=None,
              max_round=-1,
              text_spliter=None,
              model='gpt-4o-mini',
              soft_merge_prompt=None,
              calc_len=None,
              logger=None,
              ):
        """ 从一段（长）文本内容中进行信息抽取

        :param str|list[str] contents: 原文内容
            list[str]，表示本身已经有些硬分割
        :param str|list[str] extra_prompts: 提取目标信息的提示词
            str，无论递归多少轮，都用同一个提示词
            list[str]，第0轮用prompts[0]，第1轮使用prompts[1]，以此类推，如果轮次超出prompts长度，则使用prompts[-1]

        :param int segment_len: 每段文本的最大长度 （原content过长时，需要拆分成多段分开处理）
            理论上最终期望摘取的信息长度应该是不超过segment_len的一半
        :param int overlap_len: 段与段之间的重叠文本长度
        :param int max_round: 最大处理轮次，-1表示无限制，直到所有内容都被处理完
            比如max_round=1，就表示只处理一个轮次
        :param int expect_len: 最终期望的内容长度，默认是segment_len的一半
        :param func calc_len: 文本长度的算法， 默认使用len
        :param int max_workers: 提取信息时，能开的最大并发数

        :param func text_spliter: 文本分割器，默认使用均切
            均匀切分，在不超过segment_len的情况下，尽量保持每段文本长度相同
            后续要支持langchain里分割算法，觉得那个切法更泛用，虽然效率估计更低
        :param str model: 使用的模型
        :param soft_merge_prompt: 最后一轮合并后，如果已经满足要求，是否要加一段提示词，再优化下描述，避免暴力合并效果过于生硬

        :param logger: 日志记录器，输入None表示不记录。 暂未实装，后续可能要用来评估token长度。
        """
        # 1 预备
        if isinstance(contents, str):
            contents = [contents]

        if isinstance(extra_prompts, str):
            extra_prompts = [extra_prompts]

        expect_len = expect_len or segment_len // 2
        logger = logger or MagicMock()
        if logger is True:
            logger = loguru_logger

        calc_len = calc_len or len

        if text_spliter is None:
            def text_spliter(content, segment_len=segment_len, overlap_len=overlap_len):
                return TextSpliter.avg(content, segment_len, overlap_len)

        # 2 循环提取信息
        round_num = 0
        while True:
            # 1 判断现在的长度，是否需要精简信息提取
            last_content_len = sum([calc_len(c) for c in contents])
            if last_content_len <= expect_len:
                break

            # 2 切分文本
            contents2 = []
            for content in contents:
                contents2.extend(text_spliter(content))

            # 3 单个提取的接口
            def extract_info(i, content, prompt):
                chat = Chat(model=model)
                chat.add_system_prompt(prompt)
                summary = chat.query(content)
                logger.debug(f'>>> 第{i}块原始内容：\n{content}\n\n>>> 提取信息：{summary}\n\n')
                return summary

            # 4 (并发)提取信息
            prompt_idx = min(round_num, len(extra_prompts) - 1)
            backend = 'threading' if max_workers != 1 else 'sequential'
            parallel = Parallel(n_jobs=max_workers, backend=backend, return_as='generator')
            logger.debug(f'>>> 提取信息轮次：{round_num}，使用提示词：\n{extra_prompts[prompt_idx]}')
            tasks = [delayed(extract_info)(i, c, extra_prompts[prompt_idx]) for i, c in enumerate(contents2, start=1)]
            contents = list(tqdm(parallel(tasks), total=len(contents2),
                                 desc=f'round {round_num} 提取信息', disable=isinstance(logger, MagicMock)))

            # 5 合并数据
            contents = ['\n\n'.join(contents)]
            round_num += 1

            if max_round > 0 and round_num >= max_round:
                break

        # 3 提取结果
        res = '\n\n'.join(contents)
        if soft_merge_prompt:
            chat = Chat(model=model)
            chat.add_system_prompt(soft_merge_prompt)
            res = chat.query(res)

        return res

    @classmethod
    def common(cls, content, extra_prompts=None, **kwargs):
        """ 通用的信息压缩 """
        return cls.basic(content,
                         extra_prompts or xlprompts.compress_content_common,
                         **kwargs)

    @classmethod
    def with_query(cls, content, query=None, extra_prompts=None, **kwargs):
        """ 通用的信息压缩，根据用户的问题，进行信息压缩
        但如果没传入query参数，也会自动改回common机制处理
        """
        if not query:
            return cls.common(content, extra_prompts, **kwargs)

        return cls.basic(content,
                         extra_prompts or xlprompts.compress_content_with_query.render(query=query),
                         **kwargs)

    @classmethod
    def qq_message(cls, content, extra_prompts=None, **kwargs):
        """ QQ聊天记录整理 """
        return cls.basic(content,
                         extra_prompts or xlprompts.compress_qq_message,
                         **kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 10:51

from pyxllib.prog.newbie import round_int


class StrDecorator:
    """将函数的返回值字符串化，仅调用朴素的str字符串化

    装饰器开发可参考： https://mp.weixin.qq.com/s/Om98PpncG52Ba1ZQ8NIjLA
    """

    def __init__(self, func):
        self.func = func  # 使用self.func可以索引回原始函数名称
        self.last_raw_res = None  # last raw result，上一次执行函数的原始结果

    def __call__(self, *args, **kwargs):
        self.last_raw_res = self.func(*args, **kwargs)
        return str(self.last_raw_res)


class PrintDecorator:
    """将函数返回结果直接输出"""

    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        s = self.func(*args, **kwargs)
        print(s)
        return s  # 输出后仍然会返回原函数运行值


def binary_cut_str(s, fmt='0'):
    """180801坤泽：“二分”切割字符串
    :param s: 要截取的全字符串
    :param fmt: 截取格式，本来是想只支持0、1的，后来想想支持23456789也行
        0：左边一半
        1：右边的1/2
        2：右边的1/3
        3：右边的1/4
        ...
        9：右边的1/10
    :return: 截取后的字符串

    >>> binary_cut_str('1234', '0')
    '12'
    >>> binary_cut_str('1234', '1')
    '34'
    >>> binary_cut_str('1234', '10')
    '3'
    >>> binary_cut_str('123456789', '20')
    '7'
    >>> binary_cut_str('123456789', '210')  # 向下取整，'21'获得了9，然后'0'取到空字符串
    ''
    """
    for t in fmt:
        t = int(t)
        n = len(s) // (1 + max(1, t))
        if t == 0:
            s = s[:n]
        else:
            s = s[(len(s) - n):]
    return s


def digits2roman(d):
    """
    >>> digits2roman(2)
    'Ⅱ'
    >>> digits2roman(12)
    'Ⅻ'
    """
    rmn = '~ⅠⅡⅢⅣⅤⅥⅦⅧⅨⅩⅪⅫ'  # roman数字number的缩写

    d = int(d)  # 确保是整数类型
    if d <= 12:
        return rmn[d]
    else:
        raise NotImplementedError


def roman2digits(d):
    """
    >>> roman2digits('Ⅱ')
    2
    >>> roman2digits('Ⅻ')
    12
    """
    rmn = '~ⅠⅡⅢⅣⅤⅥⅦⅧⅨⅩⅪⅫ'
    if d in rmn:
        return rmn.index(d)
    else:
        raise NotImplemented


def digits2circlednumber(d):
    d = int(d)
    if 0 < d <= 20:
        return '①②③④⑤⑥⑦⑧⑨⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳'[d - 1]
    else:
        raise NotImplemented


def circlednumber2digits(d):
    t = '①②③④⑤⑥⑦⑧⑨⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳'
    if d in t:
        return t.index(d) + 1
    else:
        raise NotImplemented


def endswith(s, tags):
    """除了模拟str.endswith方法，输入的tag也可以是可迭代对象

    >>> endswith('a.dvi', ('.log', '.aux', '.dvi', 'busy'))
    True
    """
    if isinstance(tags, str):
        return s.endswith(tags)
    elif isinstance(tags, (list, tuple)):
        for t in tags:
            if s.endswith(t):
                return True
    else:
        raise TypeError
    return False


def xldictstr(d, key_value_delimit='=', item_delimit=' '):
    """将一个字典转成字符串"""
    res = []
    for k, v in d.items():
        res.append(str(k) + key_value_delimit + str(v).replace('\n', r'\n'))
    res = item_delimit.join(res)
    return res


def findnth(haystack, needle, n):
    """https://stackoverflow.com/questions/1883980/find-the-nth-occurrence-of-substring-in-a-string"""
    if n < 0:
        n += haystack.count(needle)
    if n < 0:
        return -1

    parts = haystack.split(needle, n + 1)
    if len(parts) <= n + 1:
        return -1
    return len(haystack) - len(parts[-1]) - len(needle)


def refine_digits_set(digits):
    """美化连续数字的输出效果

    >>> refine_digits_set([210, 207, 207, 208, 211, 212])
    '207,208,210-212'
    """
    arr = sorted(list(set(digits)))  # 去重
    n = len(arr)
    res = ''
    i = 0
    while i < n:
        j = i + 2
        if j < n and arr[i] + 2 == arr[j]:
            while j < n and arr[j] - arr[i] == j - i:
                j += 1
            j = j if j < n else n - 1
            res += str(arr[i]) + '-' + str(arr[j]) + ','
            i = j + 1
        else:
            res += str(arr[i]) + ','
            i += 1
    return res[:-1]  # -1是去掉最后一个','


def del_tail_newline(s):
    """删除末尾的换行"""
    if len(s) > 1 and s[-1] == '\n':
        s = s[:-1]
    return s


def bracket_match(s, idx):
    """括号匹配位置
    这里以{、}为例，注意也要适用于'[]', '()'
    >>> bracket_match('{123}', 0)
    4
    >>> bracket_match('0{23{5}}89', 1)
    7
    >>> bracket_match('0{23{5}}89', 7)
    1
    >>> bracket_match('0{23{5}78', 1) is None
    True
    >>> bracket_match('0{23{5}78', 20) is None
    True
    >>> bracket_match('0[2[4]{7}]01', 9)
    1
    >>> bracket_match('0{[34{6}89}', -4)
    5
    """
    key = '{[(<>)]}'
    try:
        if idx < 0:
            idx += len(s)
        ch1 = s[idx]
        idx1 = key.index(ch1)
    except ValueError:  # 找不到ch1
        return None
    except IndexError:  # 下标越界，表示没有匹配到右括号
        return None
    idx2 = len(key) - idx1 - 1
    ch2 = key[idx2]
    step = 1 if idx2 > idx1 else -1
    cnt = 1
    i = idx + step
    if i < 0:
        i += len(s)
    while 0 <= i < len(s):
        if s[i] == ch1:
            cnt += 1
        elif s[i] == ch2:
            cnt -= 1
        if cnt == 0:
            return i
        i += step
    return None


def bracket_match2(s, idx):
    r"""与“bracket_match”相比，会考虑"\{"转义字符的影响

    >>> bracket_match2('a{b{}b}c', 1)
    6
    >>> bracket_match2('a{b{\}b}c}d', 1)
    9
    """
    key = '{[(<>)]}'
    try:
        if idx < 0:
            idx += len(s)
        ch1 = s[idx]
        idx1 = key.index(ch1)
    except ValueError:  # 找不到ch1
        return None
    except IndexError:  # 下标越界，表示没有匹配到右括号
        return None
    idx2 = len(key) - idx1 - 1
    ch2 = key[idx2]
    step = 1 if idx2 > idx1 else -1
    cnt = 1
    i = idx + step
    if i < 0:
        i += len(s)
    while 0 <= i < len(s):
        if i and s[i - 1] == '\\':
            pass
        elif s[i] == ch1:
            cnt += 1
        elif s[i] == ch2:
            cnt -= 1
        if cnt == 0:
            return i
        i += step
    return None


def latexstrip(s):
    """latex版的strip"""
    return s.strip('\t\n ~')


def add_quote(s):
    return f'"{s}"'


def fold_dict(d, m=5):
    """ 将字典折叠为更紧凑的排版格式

    :param d: 一个字典对象
    :param m: 按照每行放m个元素重排
    :return: 重排后的字典内容
    """
    vals = [f"'{k}': {v}" for k, v in d.items()]
    line = [', '.join(vals[i:i + 5]) for i in range(0, len(vals), m)]
    return '{' + ',\n'.join(line) + '}'


def remove_prefix(original_string, prefix):
    if original_string.startswith(prefix):
        return original_string[len(prefix):]
    return original_string


def remove_suffix(original_string, suffix):
    if original_string.endswith(suffix):
        return original_string[:-len(suffix)]
    return original_string



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/06/09 16:26

"""
针对PostgreSQL封装的工具
"""

import sys

from pyxllib.file.specialist import XlPath

import io
from collections import Counter
import json
import json
import textwrap
import datetime
import re

from tqdm import tqdm

import psycopg
import psycopg.rows

from pyxllib.prog.newbie import round_int, human_readable_number
from pyxllib.prog.pupil import utc_now, utc_timestamp, is_valid_identifier
from pyxllib.prog.specialist import XlOsEnv
from pyxllib.algo.pupil import ValuesStat2
from pyxllib.file.specialist import get_etag, StreamJsonlWriter
from pyxllib.data.sqlite import SqlBase, SqlBuilder


class Connection(psycopg.Connection, SqlBase):

    def __init__(self, *args, **kwargs):
        psycopg.Connection.__init__(self, *args, **kwargs)
        SqlBase.__init__(self, *args, **kwargs)

    def __del__(self):
        self.close()

    def __1_库(self):
        pass

    def get_db_activities(self, datname=None):
        """
        检索当前数据库的活动信息。

        :param datname: 这个字段理论上应该要能自动检测出来才对，但这会急着没空钻研，先手动输入吧
        """
        sql = SqlBuilder('pg_stat_activity')
        sql.select('pid', 'datname', 'usename', 'state', 'query', 'age(now(), query_start) AS "query_age"')
        sql.where("state = 'active'")
        if datname:
            sql.where(f"datname = '{datname}'")
        return self.exec2dict(sql.build_select()).fetchall()

    def __2_表格(self):
        pass

    def get_table_names(self):
        # cmd = "SELECT table_name FROM information_schema.tables WHERE table_schema='public' AND table_type='BASE TABLE'"
        cmd = "SELECT tablename FROM pg_tables WHERE schemaname='public'"
        return [x[0] for x in self.execute(cmd)]

    def has_table(self, table_name, schemaname='public'):
        query = ["SELECT EXISTS (SELECT FROM pg_tables ",
                 f"WHERE schemaname='{schemaname}' AND tablename='{table_name}')"]
        res = self.execute(' '.join(query))
        return res.fetchone()[0]

    def get_column_names(self, table_name):
        """ 【查】表格有哪些字段
        """
        cmd = f"SELECT column_name FROM information_schema.columns WHERE table_name='{table_name}'"
        return self.exec2col(cmd)

    def ensure_column(self, table_name, col_name, *args, comment=None, **kwargs):
        super(Connection, self).ensure_column(table_name, col_name, *args, **kwargs)
        if comment:
            self.set_column_comment(table_name, col_name, comment)

    def set_column_comment(self, table_name, col_name, comment):
        # 这里comment按%s填充会报错。好像是psycopg库的问题，这种情况因为没有表格字段参照类型，会不知道要把comment转成字符串。
        # 然后使用py的f字符串填充的话，要避免comment有引号等特殊字符，就用了特殊的转义标记$zyzf$
        # 暂时找不到更优雅的方式~~ 而且这个问题，可能以后其他特殊的sql语句也会遇到。。。
        self.execute(f"COMMENT ON COLUMN {table_name}.{col_name} IS $zyzf${comment}$zyzf$")
        self.commit()

    def reset_table_item_id(self, table_name, item_id_name=None, counter_name=None):
        """ 重置表格的数据的id值，也重置计数器

        :param item_id_name: 表格的自增id字段名，有一套我自己风格的自动推算算法
        :param counter_name: 计数器字段名，如果有的话，也会重置
        """
        # 1 重置数据的编号
        if item_id_name is None:
            m = re.match(r'(.+?)(_table)?$', table_name)
            item_id_name = m.group(1) + '_id'

        sql = f"""WITH cte AS (
    SELECT {item_id_name}, ROW_NUMBER() OVER (ORDER BY {item_id_name}) AS new_{item_id_name}
    FROM {table_name}
)
UPDATE {table_name}
SET {item_id_name} = cte.new_{item_id_name}
FROM cte
WHERE {table_name}.{item_id_name} = cte.{item_id_name}"""
        self.execute(sql)  # todo 这种sql写法好像偶尔有bug会出问题
        self.commit()

        # 2 重置计数器
        if counter_name is None:
            counter_name = f'{table_name}_{item_id_name}_seq'

        # 找到目前最大的id值
        max_id = self.exec2one(f'SELECT MAX({item_id_name}) FROM {table_name}')
        # self.execute(f'ALTER SEQUENCE {counter_name} RESTART WITH {max_id + 1}')
        # 检查序列是否存在，如果不存在则创建序列，然后重置序列
        # 检查序列是否存在，如果不存在则创建序列，并将其关联到指定的表和字段
        self.execute(f"""
DO $$
BEGIN
    IF NOT EXISTS (SELECT FROM pg_class WHERE relkind = 'S' AND relname = '{counter_name}') THEN
        EXECUTE format('CREATE SEQUENCE %I', '{counter_name}');
        EXECUTE format('ALTER TABLE %I ALTER COLUMN %I SET DEFAULT nextval(''%I'')', '{table_name}', '{item_id_name}', '{counter_name}');
    END IF;
    EXECUTE format('ALTER SEQUENCE %I RESTART WITH %s', '{counter_name}', {max_id + 1});
END $$;
""")
        self.commit()

    def __3_execute(self):
        pass

    def exec_nametuple(self, *args, **kwargs):
        cur = self.cursor(row_factory=psycopg.rows.namedtuple_row)
        data = cur.execute(*args, **kwargs)
        # cur.close()
        return data

    def exec2dict(self, *args, **kwargs):
        cur = self.cursor(row_factory=psycopg.rows.dict_row)
        data = cur.execute(*args, **kwargs)
        # cur.close()
        return data

    def exec2dict_batch(self, sql, batch_size=1000, use_offset=None, **kwargs):
        """ 分批返回数据的版本

        :param use_offset: 是否使用offset分页，会根据sql中是否含有where自动判断，但有时候最好明确指定以防错误
            如果外部sql每次操作，会改变数据库的情况，导致sql的where规则虽然没变，但是数据本身发生变化，则offset应该要关闭
                每次取对应的满足条件的数据即可
                这种情况，也需要本函数内部主动执行commit_all的
            否则，只是一种遍历查询，没有where或者where获取的数据情况是不会变化的，则要使用offset
        :return:
            第1个值，是一个迭代器，看起来仍然能一条一条返回，实际后台是按照batch_size打包获取的
            第2个值，是数据总数
        """
        if not isinstance(sql, SqlBuilder):
            raise ValueError('暂时只能搭配SQLBuilder使用')

        if use_offset is None:
            use_offset = not sql._where

        num = self.exec2one(sql.build_count())
        offset = 0

        def yield_row():
            nonlocal offset
            while True:
                sql2 = sql.copy()
                if not use_offset:  # 如果不使用offset，那么缓存的sql操作需要全部提交，确保数据都更新后，再提取数据
                    self.commit_all()
                sql2.limit(batch_size, offset)
                rows = self.exec2dict(sql2.build_select(), **kwargs).fetchall()
                if use_offset:
                    offset += len(rows)
                if not rows:
                    break
                yield from rows

        return yield_row(), num

    exec_dict = exec2dict

    def __4_数据类型(self):
        pass

    @classmethod
    def cvt_type(cls, val):
        if isinstance(val, (dict, list)):
            val = json.dumps(val, ensure_ascii=False, default=str)
        # 注意list数组类型读、写都会自动适配py
        return val

    @classmethod
    def autotype(cls, val):
        if isinstance(val, str):
            return 'text'
        elif isinstance(val, int):
            return 'int4'  # int2、int8
        elif isinstance(val, bool):
            return 'boolean'
        elif isinstance(val, float):
            return 'float4'
        elif isinstance(val, (dict, list)):
            return 'jsonb'
        elif isinstance(val, datetime.datetime):
            return 'timestamp'
        else:  # 其他list等类型，可以用json.dumps或str转文本存储
            return 'text'

    def __5_增删改查(self):
        pass

    def insert_row(self, table_name, cols, *, on_conflict='DO NOTHING', commit=False):
        """ 【增】插入新数据

        :param dict cols: 用字典表示的要插入的值
        :param on_conflict: 如果已存在的处理策略
            DO NOTHING，跳过不处理
            REPLACE，这是个特殊标记，会转为主键冲突也仍然更新所有值
            (id) DO UPDATE SET host_name=EXCLUDED.host_name, nick_name='abc'
                也可以写复杂的处理算法规则，详见 http://postgres.cn/docs/12/sql-insert.html
                比如这里是插入的id重复的话，就把host_name替换掉，还可以指定nick_name替换为'abc'
                注意前面的(id)是必须要输入的

        注意：有个常见需求，是想插入后返回对应的id，但是这样就需要知道这张表自增的id字段名
            以及还是很难获得插入后的id值，可以默认刚插入的id是最大的，但是这样并不安全，有风险
            建议还是外部自己先计算全表最大的id值，自己实现自增，就能知道插入的这条数据的id了
        """
        ks = ','.join(cols.keys())
        vs = ','.join(['%s'] * (len(cols.keys())))
        query = f'INSERT INTO {table_name}({ks}) VALUES ({vs})'
        params = self.cvt_types(cols.values())

        if on_conflict == 'REPLACE':
            on_conflict = f'ON CONFLICT ON CONSTRAINT {table_name}_pkey DO UPDATE SET ' + \
                          ','.join([f'{k}=EXCLUDED.{k}' for k in cols.keys()])
        else:
            on_conflict = f'ON CONFLICT {on_conflict}'
        query += f' {on_conflict}'

        self.commit_base(commit, query, params)

    def insert_rows(self, table_name, keys, ls, *, on_conflict='DO NOTHING', commit=False):
        """ 【增】插入新数据

        :param str keys: 要插入的字段名，一个字符串，逗号,隔开属性值
        :param list[list] ls: n行m列的数组

        >> con.insert_rows('hosts2', 'id,host_name,nick_name', [[1, 'test5', 'dcba'], [11, 'test', 'aabb']])
        """
        n, m = len(ls), len(keys.split(','))
        vs = ','.join(['%s'] * m)
        query = f'INSERT INTO {table_name}({keys}) VALUES ' + ','.join([f'({vs})'] * n)
        params = []
        for cols in ls:
            params += self.cvt_types(cols)

        if on_conflict == 'REPLACE':
            on_conflict = f'ON CONFLICT ON CONSTRAINT {table_name}_pkey DO UPDATE SET ' + \
                          ','.join([f'{k.strip()}=EXCLUDED.{k.strip()}' for k in keys.split(',')])
        else:
            on_conflict = f'ON CONFLICT {on_conflict}'
        query += f' {on_conflict}'

        self.commit_base(commit, query, params)

    def __6_高级统计(self):
        pass

    def get_column_valuesstat(self, table_name, column, filter_condition=None,
                              percentile_count=5,
                              by_data=False, data_type=None):
        """ 获得指定表格的某个字段的统计特征ValuesStat2对象

        :param table_name: 表名
        :param column: 用于计算统计数据的字段名
            不一定是标准的字段名
        :param percentile_count: 分位数的数量，例如 3 表示只计算中位数
        :param by_data: 是否获得原始数据
            默认只获得统计特征，不获得原始数据
        """

        def init_from_db_data():
            sql = SqlBuilder(table_name)
            if filter_condition:
                sql.where(filter_condition)
            values = self.exec2col(sql.build_select(column))
            if data_type == 'numeric':
                values = [x and float(x) for x in values]
            vs = ValuesStat2(raw_values=values, data_type=data_type)

            if data_type == 'text' and is_valid_identifier(column):
                vs0 = self.get_column_valuesstat(table_name, column, filter_condition=filter_condition,
                                                 percentile_count=percentile_count, by_data=False)
                vs.n = vs0.n
                vs.dist = vs0.dist

            return vs

        def init_from_db():
            # 1 构建基础的 SQL 查询
            sql = SqlBuilder(table_name)
            sql.select("COUNT(*) AS total_count")
            sql.select(f"COUNT({column}) AS non_null_count")
            sql.select(f"MIN({column}) AS min_value")
            sql.select(f"MAX({column}) AS max_value")
            if data_type and 'timestamp' in data_type:
                percentile_type = 'PERCENTILE_DISC'
                # todo 其实时间类也可以"泛化"一种平均值、标准差算法的，这需要获取全量数据，然后自己计算
            elif data_type == 'text':
                percentile_type = 'PERCENTILE_DISC'
            else:  # 默认是正常的数值类型
                sql.select(f"SUM({column}) AS total_sum")
                sql.select(f"AVG({column}) AS average")
                sql.select(f"STDDEV({column}) AS standard_deviation")
                percentile_type = 'PERCENTILE_CONT'

            percentiles = []
            # 根据分位点的数量动态添加分位数计算
            if percentile_count > 2:
                step = 1 / (percentile_count - 1)
                percentiles = [(i * step) for i in range(1, percentile_count - 1)]
                for p in percentiles:
                    sql.select(f"{percentile_type}({p:.2f}) WITHIN GROUP (ORDER BY {column}) "
                               f"AS percentile_{int(p * 100)}")

            if filter_condition:
                sql.where(filter_condition)

            row = self.exec2dict(sql.build_select()).fetchone()

            # 2 统计展示
            x = ValuesStat2(data_type=data_type)
            x.raw_n = row['total_count']
            x.n = row['non_null_count']
            if not x.n:
                return x

            x.sum = row.get('total_sum', None)
            x.mean = row.get('average', None)
            x.std = row.get('standard_deviation', None)

            # 如果计算了分位数，填充相应属性
            x.dist = [row['min_value']] + [row[f"percentile_{int(p * 100)}"] for p in percentiles] + [row['max_value']]
            if data_type == 'numeric':
                x.dist = [float(x) for x in x.dist]

            return x

        data_type = data_type or self.get_column_data_type(table_name, column)

        # 如果不是标准的列名，强制获取数据
        if not is_valid_identifier(column):
            by_data = True

        if by_data:
            vs = init_from_db_data()
        else:
            vs = init_from_db()

        return vs

    def export_jsonl(self, file_path, table_name, key_col=None, batch_size=1000, print_mode=0):
        """ 将某个表导出为本地jsonl文件

        :param str|SqlBuilder table_name: 表名
            支持传入SqlBuilder对象，这样可以更灵活的控制导出的数据规则
        :param file_path: 导出的文件路径
        :param batch_size: 每次读取的行数和保存的行数
        :param key_col: 作为主键的列名，如果有的话，会自动去重
            强烈推荐要设置
            实际不一定要用主键，只要是有顺序值的列就行

        todo 暴力最简单的版本不难写，我纠结的是缓存机制，还有bytes类型数据会有点大等问题
            还需要先支持一个通用的缓存写文件功能
        """
        # 1 sql
        if isinstance(table_name, str):
            sql = SqlBuilder(table_name)
            sql.select('*')
        else:
            sql = table_name
            m = re.search(r'FROM (\w+)', sql.build_select())
            table_name = m.group(1) if m else 'table'
        assert isinstance(sql, SqlBuilder)

        file_path = XlPath(file_path)
        if key_col:
            sql.order_by(key_col)
            if file_path.is_file():
                # 读取现有数据，找出主键最大值
                data = file_path.read_jsonl(batch_size=1000)
                if data:
                    max_val = max([x[key_col] for x in data]) if data else None
                    if max_val is not None:
                        sql.where(f'{key_col} > {max_val}')

        # 2 获取数据
        file = StreamJsonlWriter(file_path, batch_size=batch_size)  # 流式存储
        rows, total = self.exec2dict_batch(sql, batch_size=batch_size, use_offset=True)
        for row in tqdm(rows, total=total, desc=f'从{table_name}表导出数据', disable=not print_mode):
            file.append_line(row)
        file.flush()

    def check_db_tables_size(self, db_name=None):
        """ 查看指定数据下所有表格的大小 """
        from datetime import datetime
        import pandas as pd

        if db_name is None:
            # 使用sql获取当前self所在数据库
            db_name = self.exec2one("SELECT current_database()")

        data = []
        tables = self.exec2col("SELECT table_name FROM information_schema.tables WHERE table_schema='public'")
        for table_name in tables:
            row = {
                'database': db_name,
                'table_name': table_name,
            }
            sz = self.exec2one(f"SELECT pg_total_relation_size('public.{table_name}')")
            if not sz:
                continue
            lines = self.exec2one(f"SELECT COUNT(*) FROM {table_name}")
            row['size'], row['lines'] = sz, lines
            row['readable_size'] = human_readable_number(sz, 'KB')
            row['perline_size'] = human_readable_number(sz / lines, 'KB') if lines else -1
            row['update_time'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')
            data.append(row)

        df = pd.DataFrame.from_dict(data)
        if len(df):
            df.sort_values(['size'], ascending=False, inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df

    def check_multi_db_size(self, db_list):
        """ 这个功能一般要用postgres账号，才有权限处理所有数据库 """
        from datetime import datetime
        import pandas as pd

        data = []
        for db in db_list:
            row = {
                'name': db,
            }
            sz = self.exec2one(f"SELECT pg_database_size('{db}')")
            row['size'] = sz
            row['readable_size'] = human_readable_number(sz, 'KB')
            row['update_time'] = datetime.now().strftime('%Y-%m-%d %H:%M:%S')

            data.append(row)

        df = pd.DataFrame.from_dict(data)
        df.sort_values(['size'], ascending=False, inplace=True)
        df.reset_index(drop=True, inplace=True)
        return df


"""
【关于为什么XlprDb要和pglib合一个文件】
好处
1、代码集中管理，方便开发效率

坏处 - 对治
1、导入代码多，效率低
    - 导入代码其实很快的，不差这一点点
2、仅用简单功能的时候，会引入多余复杂功能
    - 如果单拉一个xlprdb.py文件，依然不能头部import复杂库，因为XlprDb也存在需要精简使用的场景
        既然都要有精简的部分，干脆都把代码弄到一起得了
        然后复杂包都在成员函数里单独import
"""


class XlprDb(Connection):
    """ xlpr统一集中管理的一个数据库

    为了一些基础的数据库功能操作干净，尽量不要在全局导入过于复杂的包，可以在每个特定功能里导特定包
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seckey = ''

    @classmethod
    def set_conninfo(cls, conninfo, seckey=''):
        """ 提前将登录信息加密写到环境变量中，这样代码中就没有明文的账号密码信息

        :param conninfo:
        :param seckey: 如果要获得数据库里较重要的密码等信息，需要配置key值，否则默认可以不设

        使用后，需要重启IDE重新加载环境变量
        并且本句明文代码需要删除
        """
        # TODO 目前只设一个账号，后续可以扩展支持多个账号指定配置
        # conninfo = 'postgresql://postgres:yourpassword@172.16.170.110/xlpr'
        return XlOsEnv.persist_set('XlprDbAccount', {'conninfo': conninfo, 'seckey': seckey}, encoding=True)

    @classmethod
    def connect(cls, conninfo='', seckey='', *,
                autocommit=False, row_factory=None, context=None, **kwargs) -> 'XlprDb':
        """ 因为要标记 -> XlprDb，IDE才会识别出类别，有自动补全功能
        但在类中写@classmethod，无法标记 -> XlprDb，所以就放外面单独写一个方法了
        """
        d = XlOsEnv.get('XlprDbAccount', decoding=True)
        if conninfo == '':
            conninfo = d['conninfo']
        if seckey == '' and isinstance(d, dict) and 'seckey' in d:
            seckey = d['seckey']
        # 注意这里获取的是XlprDb类型
        con = super(XlprDb, cls).connect(conninfo, autocommit=autocommit, row_factory=row_factory, context=context,
                                         **kwargs)
        con.seckey = seckey
        return con

    @classmethod
    def connect2(cls, name, passwd, database=None, ip_list=None, **kwargs):
        """ 简化的登录方式，并且自带优先尝试局域网连接，再公网连接

        一般用在jupyter等对启动速度没有太高要求的场合，因为不断尝试localhost等需要耗费不少时间
        """
        if database is None:
            database = name

        xldb = None
        if ip_list is None:
            ip_list = ['localhost', '172.16.170.136', 'xmutpriu.com']
        for ip in ip_list:
            try:
                xldb = cls.connect(f'postgresql://{name}:{passwd}@{ip}/{database}', **kwargs)
                break
            except psycopg.OperationalError:
                pass

        return xldb

    def __1_hosts相关数据表操作(self):
        pass

    def update_host(self, host_name, accounts=None, **kwargs):
        """ 更新一台服务器的信息

        :param dict accounts: 账号信息，记得要列全，比如
            {'root': '123456', 'chenkunze': '654321'}
        """
        if not self.execute('SELECT EXISTS (SELECT FROM hosts WHERE host_name=%s)', (host_name,)).fetchone()[0]:
            self.insert_row('hosts', {'host_name': host_name})
        if kwargs:
            self.update_row('hosts', kwargs, {'host_name': host_name})
        if accounts:
            self.execute('UPDATE hosts SET accounts=pgp_sym_encrypt(%s, %s) WHERE host_name=%s',
                         (json.dumps(accounts, ensure_ascii=False), self.seckey, host_name))
        self.commit()

    def set_host_account(self, host_name, user_name, passwd):
        """ 设置某台服务器的一个账号密码

        >> xldb.set_host_account('titan2', 'chenkunze', '123456')

        """
        # 读取旧的账号密码字典数据
        d = self.execute("SELECT pgp_sym_decrypt(accounts, %s)::jsonb FROM hosts WHERE host_name=%s",
                         (self.seckey, host_name)).fetchone()[0]
        # 修改字典数据
        d[user_name] = str(passwd)
        # 将新的字典数据写回数据库
        self.execute('UPDATE hosts SET accounts=pgp_sym_encrypt(%s, %s) WHERE host_name=%s',
                     (json.dumps(d, ensure_ascii=False), self.seckey, host_name))
        self.commit()

    def login_ssh(self, host_name, user_name, map_path=None, **kwargs) -> 'XlSSHClient':
        """ 通过数据库里的服务器数据记录，直接登录服务器 """
        from pyxllib.ext.unixlib import XlSSHClient

        if host_name.startswith('g_'):
            host_ip = self.execute("SELECT host_ip FROM hosts WHERE host_name='xlpr0'").fetchone()[0]
            pw, port = self.execute('SELECT (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text, frpc_port'
                                    ' FROM hosts WHERE host_name=%s',
                                    (self.seckey, user_name, host_name[2:])).fetchone()
        else:
            port = 22
            host_ip, pw = self.execute('SELECT host_ip, (pgp_sym_decrypt(accounts, %s)::jsonb)[%s]::text'
                                       ' FROM hosts WHERE host_name=%s',
                                       (self.seckey, user_name, host_name)).fetchone()

        if map_path is None:
            if sys.platform == 'win32':
                map_path = {'C:/': '/'}
            else:
                map_path = {'/': '/'}

        return XlSSHClient(host_ip, user_name, pw[1:-1], port=port, map_path=map_path, **kwargs)

    def __2_xlapi相关数据表操作(self):
        """
        files，存储二进制文件的表
            etag，文件、二进制对应的etag值
            meta，可以存储不同数据类型一些特殊的属性，比如图片可以存储高、宽
                但因为用户使用中，为了提高速度，以及减少PIL等依赖，执行中不做计算
                可以其他途径使用定期脚本自动处理
        xlapi，底层api调用的记录统计
            input，所有输入的参数
                mode，使用的算法接口
                etag，涉及到大文件数据的，以打包后的文件的etag做记录
            output，运行结果
            elapse_ms，调用函数的用时，不含后处理转换格式的时间
        xlserver，服务端记录的使用情况
        """
        pass

    def insert_row2files(self, buffer, *, etag=None, **kwargs):
        """

        为了运算效率考虑，除了etag需要用于去重，是必填字段
        其他字段默认先不填充计算
        """
        # 1 已经有的不重复存储
        if etag is None:
            etag = get_etag(buffer)

        res = self.execute('SELECT etag FROM files WHERE etag=%s', (etag,)).fetchone()
        if res:  # 已经做过记录的，不再重复记录
            return

        # 2 没有的图，做个备份
        kwargs['etag'] = etag
        kwargs['data'] = buffer
        kwargs['fsize_kb'] = round_int(len(buffer) / 1024)

        self.insert_row('files', kwargs)
        self.commit()

    def get_xlapi_record(self, **input):
        res = self.execute('SELECT id, output FROM xlapi WHERE input=%s', (self.cvt_type(input),)).fetchone()
        if res:
            _id, output = res
            output['xlapi_id'] = _id
            return output

    def insert_row2xlapi(self, input, output, elapse_ms, *, on_conflict='(input) DO NOTHING'):
        """ 往数据库记录当前操作内容

        :return: 这个函数比较特殊，需要返回插入的条目的id值
        """
        if on_conflict == 'REPLACE':
            on_conflict = "(input) DO UPDATE " \
                          "SET output=EXCLUDED.output, elapse_ms=EXCLUDED.elapse_ms, update_time=EXCLUDED.update_time"

        input = json.dumps(input, ensure_ascii=False)
        self.insert_row('xlapi', {'input': input, 'output': output,
                                  'elapse_ms': elapse_ms, 'update_time': utc_timestamp(8)},
                        on_conflict=on_conflict)
        self.commit()
        return self.execute('SELECT id FROM xlapi WHERE input=%s', (input,)).fetchone()[0]

    def insert_row2xlserver(self, request, xlapi_id=0, **kwargs):
        kw = {'remote_addr': request.headers.get('X-Real-IP', request.remote_addr),
              'route': '/'.join(request.base_url.split('/')[3:]),
              'update_time': utc_timestamp(8),
              'xlapi_id': xlapi_id}
        kw.update(kwargs)
        print(kw)  # 监控谁在用api
        self.insert_row('xlserver', kw, commit=True)

    def __3_host_trace相关可视化(self):
        """ TODO dbview 改名 host_trace """
        pass

    def __dbtool(self):
        pass

    def record_host_usage(self, cpu=True, gpu=True, disk=False):
        """ 记录服务器各种状况，存储到PG数据库

        TODO 并行处理
        TODO 功能还可以增加：gpu显卡温度、硬盘读写速率检查、网络上传下载带宽
        """
        # 1 服务器列表
        host_names = self.exec2col('SELECT host_name FROM hosts WHERE id > 1 ORDER BY id')
        host_cpu_gb = {h: v for h, v in self.execute('SELECT host_name, cpu_gb FROM hosts')}

        # 2 去所有服务器取使用情况
        for i, host_name in enumerate(host_names, start=1):
            print('-' * 20, i, host_name, '-' * 20)
            try:
                ssh = self.login_ssh(host_name, 'root', relogin=5, relogin_interval=0.2)
                status = {}
                if cpu:
                    data = ssh.check_cpu_usage(print_mode=True)
                    status['cpu'] = {k: v[0] for k, v in data.items()}
                    status['cpu_memory'] = {k: round(v[1] * host_cpu_gb[host_name] / 100, 2) for k, v in data.items()}
                if gpu:
                    status['gpu_memory'] = ssh.check_gpu_usage(print_mode=True)
                if disk:
                    # 检查磁盘空间会很慢，如果超时可以跳过。（设置超时6小时）
                    status['disk_memory'] = ssh.check_disk_usage(print_mode=True, timeout=60 * 60 * 6)
            except Exception as e:
                status = {'error': f'{str(type(e))[8:-2]}: {e}'}
                print(status)

            if status:
                self.insert_row('host_trace',
                                {'host_name': host_name, 'status': status, 'update_time': utc_timestamp(8)},
                                commit=True)
            print()

    def _get_host_trace_total(self, mode, title, yaxis_name, date_trunc, recent, host_attr):
        # CREATE INDEX ON gpu_trace (update_time);  -- update_time最好建个索引
        ls = self.execute(textwrap.dedent(f"""\
        WITH cte1 AS (  -- 筛选近期内的数据，并且时间做trunc处理
            SELECT host_name, (status)['{mode}'], date_trunc('{date_trunc}', update_time) ttime
            FROM host_trace WHERE update_time > %s AND (status ? '{mode}')
        ), cte2 AS (  -- 每个时间里每个服务器的多条记录取平均
            SELECT ttime, cte1.host_name, jsonb_div(jsonb_deep_sum(status), count(*)) status
            FROM cte1 GROUP BY ttime, cte1.host_name
        )  -- 接上每台服务器显存总值，并且分组得到每个时刻总情况
        SELECT ttime, {host_attr}, jsonb_deep_sum(status)
        FROM cte2 JOIN hosts ON cte2.host_name = hosts.host_name
        GROUP BY ttime ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self._create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _get_host_trace_per_host(self, hostname, mode, title, yaxis_name, date_trunc, recent, host_attr):
        ls = self.execute(textwrap.dedent(f"""\
        WITH cte1 AS (
            SELECT (status)['{mode}'], date_trunc('{date_trunc}', update_time) ttime
            FROM host_trace WHERE host_name='{hostname}' AND update_time > %s AND (status ? '{mode}')
        ), cte2 AS (
            SELECT ttime, jsonb_div(jsonb_deep_sum(status), count(*)) status
            FROM cte1 GROUP BY ttime
        )
        SELECT ttime, {host_attr}, jsonb_deep_sum(status)
        FROM cte2 JOIN hosts ON hosts.host_name='{hostname}'
        GROUP BY ttime ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        if not ls:  # 有的服务器可能数据是空的
            return '', 0
        return self._create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _create_stack_chart(self, title, ls, *, yaxis_name=''):
        """ 创建展示表

        :param title: 表格标题
        :param list ls: n*3，第1列是时间，第2列是总值，第3列是每个用户具体的数据
        """
        from pyecharts.charts import Line

        map_user_name = {}
        for ks, v in self.execute('SELECT account_names, name FROM users'):
            for k in ks:
                map_user_name[k] = v

        # 1 计算涉及的所有用户以及使用总量
        all_users_usaged = Counter()
        last_time = None
        for x in ls:
            hours = 0 if last_time is None else ((x[0] - last_time).total_seconds() / 3600)
            last_time = x[0]
            for k, v in x[2].items():
                if k == '_total':
                    continue
                all_users_usaged[map_user_name.get(k, k)] += v * hours

        # ls里的姓名也要跟着更新
        for i, x in enumerate(ls):
            ct = Counter()
            for k, v in x[2].items():
                ct[map_user_name.get(k, k)] += v
            ls[i] = (x[0], x[1], ct)

        # 2 转图表可视化
        def to_list(values):
            return [(x[0], v) for x, v in zip(ls, values)]

        def pretty_val(v):
            return round_int(v) if v > 100 else round(v, 2)

        try:
            chart = Line()
            chart.set_title(title)
            chart.options['xAxis'][0].update({'min': ls[0][0], 'type': 'time',
                                              # 'minInterval': 3600 * 1000 * 24,
                                              'name': '时间', 'nameGap': 50, 'nameLocation': 'middle'})
            chart.options['yAxis'][0].update({'name': yaxis_name, 'nameGap': 50, 'nameLocation': 'middle'})
            # 目前是比较暴力的方法调整排版，后续要研究是不是能更自动灵活些
            chart.options['legend'][0].update({'top': '6%', 'icon': 'pin'})
            chart.options['grid'] = [{'top': 55 + len(all_users_usaged) * 4, 'containLabel': True}]
            chart.options['tooltip'].opts.update({'axisPointer': {'type': 'cross'}, 'trigger': 'item'})

            chart.add_series(f'total{pretty_val(ls[0][1]):g}', to_list([x[1] for x in ls]), areaStyle={})
            for user, usaged in all_users_usaged.most_common():
                usaged = usaged / ((ls[-1][0] - ls[0][0]).total_seconds() / 3600 + 1e-9)
                chart.add_series(f'{user}{pretty_val(usaged):g}',
                                 to_list([x[2].get(user, 0) for x in ls]),
                                 areaStyle={}, stack='Total', emphasis={'focus': 'series'})

            return '<body>' + chart.render_embed() + '</body>', sum(all_users_usaged.values())
        except Exception as e:
            return str(e), 0

    #   cdx_edit
    def _get_database_trace_total(self, title, yaxis_name, date_trunc, recent, link_name):
        ls = self.execute(textwrap.dedent(f"""\
            WITH cte1 AS(
                SELECT link_name, jsonb_each(status::jsonb) AS db_data, date_trunc('{date_trunc}', update_time) ttime
                FROM database_trace WHERE update_time > %s AND link_name = '{link_name}'
            ), cte2 AS(
                SELECT ttime, link_name, (db_data).key AS table_name, ((db_data).value->> '_total')::bigint AS total
                FROM cte1
            )
            SELECT ttime, jsonb_object_agg(table_name, total) AS aggregated_json,SUM(total) as total
            FROM cte2
            GROUP BY ttime
            ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self.database_create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def _get_database_trace_per_host(self, db, title, yaxis_name, date_trunc, recent, link_name):
        ls = self.execute(textwrap.dedent(f"""\
            WITH cte1 AS (
                SELECT link_name, jsonb_each(status::jsonb) AS db_data, date_trunc('{date_trunc}', update_time) ttime
                FROM database_trace WHERE update_time > %s AND link_name = '{link_name}'
                ), cte2 AS (
                    SELECT ttime, link_name, (db_data).key AS table_name, (db_data).value AS size_text
                    FROM cte1
                ), cte3 AS (
                    SELECT ttime, table_name, each.key AS key, each.value AS value
                    FROM cte2, jsonb_each_text(size_text) AS each(key, value)
                )
                SELECT ttime, jsonb_object_agg(key,
                    CASE
                        WHEN key = '_total' THEN NULL
                        ELSE (value::jsonb ->> 'size')::bigint  -- Handle other keys as usual
                    END
                ) FILTER (WHERE key != '_total') AS aggregated_result,  -- 确保 _total 不在 aggregated_result 中
                MAX(CASE WHEN key = '_total' THEN value::bigint ELSE NULL END) AS total  -- 单独提取 _total 的值
                FROM cte3
                WHERE (key = '_total' OR value::jsonb ? 'size')  -- Ensure that '_total' is included
                AND table_name = '{db}'
                GROUP BY ttime
                ORDER BY ttime"""), ((utc_now(8) - recent).isoformat(timespec='seconds'),)).fetchall()
        return self.database_create_stack_chart(title, ls, yaxis_name=yaxis_name)

    def database_create_stack_chart(self, title, ls, *, yaxis_name=''):
        """ 创建展示表
    
        :param title: 表格标题
        :param list ls: n*3，第1列是时间，第3列是总值，第2列是每个用户具体的数据
        """
        from pyecharts.charts import Line
        all_database_usaged = Counter()
        last_time = None
        for x in ls:
            hours = 0 if last_time is None else ((x[0] - last_time).total_seconds() / 3600)
            last_time = x[0]
            for k, v in x[1].items():
                if k == '_total':
                    continue
                all_database_usaged[k] += v * hours

        for i, x in enumerate(ls):
            ct = Counter()
            for k, v in x[1].items():
                ct[k] += v
            ls[i] = (x[0], ct, int(x[2]))

        # 2 转图表可视化
        def to_list(values):
            return [(x[0], v) for x, v in zip(ls, values)]

        def pretty_val(v):
            return round_int(v) if v > 100 else round(v, 2)

        chart = Line()
        chart.set_title(title)
        chart.options['xAxis'][0].update({'min': ls[0][0], 'type': 'time',
                                          # 'minInterval': 3600 * 1000 * 24,
                                          'name': '时间', 'nameGap': 50, 'nameLocation': 'middle'})
        chart.options['yAxis'][0].update({'name': yaxis_name, 'nameGap': 50, 'nameLocation': 'middle'})
        # 目前是比较暴力的方法调整排版，后续要研究是不是能更自动灵活些
        chart.options['legend'][0].update({'top': '6%', 'icon': 'pin'})
        chart.options['grid'] = [{'top': 55 + len(all_database_usaged) * 4, 'containLabel': True}]
        chart.options['tooltip'].opts.update({'axisPointer': {'type': 'cross'}, 'trigger': 'item'})

        # chart.add_series(f'total {pretty_val(ls[0][2] / 1024 / 1024 / 1024):g}',
        #                 to_list([x[2] / 1024 / 1024 / 1024 for x in ls]), areaStyle={})
        for database, usaged in all_database_usaged.most_common():
            usaged = usaged / ((ls[-1][0] - ls[0][0]).total_seconds() / 3600 + 1e-9)
            chart.add_series(f'{database} {pretty_val(usaged / 1024 / 1024 / 1024):g}',
                             to_list([x[1].get(database, 0) / 1024 / 1024 / 1024 for x in ls]),
                             areaStyle={}, stack='Total', emphasis={'focus': 'series'})
        return '<body>' + chart.render_embed() + '</body>'

    def dbview_xldb1_memory(self, recent=datetime.timedelta(days=180), date_trunc='day'):
        from pyxllib.data.echarts import render_echart_html

        db_list = ['stdata', 'xlpr', 'st', 'ckz']
        args = ['数据库大小(GB)', date_trunc, recent, 'xldb1']
        htmltexts = []

        res = self._get_database_trace_total('xldb1数据库使用近况', *args)
        htmltexts.append(res)

        data_stats = []
        for idx, db in enumerate(db_list, start=1):
            data_stats.append(self._get_database_trace_per_host(db, f'{db}', *args))
        htmltexts += data_stats

        self.commit()
        h = render_echart_html('database_cdx', body='<br/>'.join(htmltexts))
        return h

    def dbview_xldb2_memory(self, recent=datetime.timedelta(days=180), date_trunc='day'):
        from pyxllib.data.echarts import render_echart_html

        db_list = ['ragdata', 'kq5034']
        args = ['数据库大小(GB)', date_trunc, recent, 'xldb2']
        htmltexts = []

        res = self._get_database_trace_total('xldb2数据库使用近况', *args)
        htmltexts.append(res)

        data_stats = []
        for idx, db in enumerate(db_list, start=1):
            data_stats.append(self._get_database_trace_per_host(db, f'{db}', *args))
        htmltexts += data_stats

        self.commit()
        h = render_echart_html('database_cdx', body='<br/>'.join(htmltexts))
        return h

    def dbview_cpu(self, recent=datetime.timedelta(days=1), date_trunc='hour'):
        from pyxllib.data.echarts import render_echart_html

        args = ['CPU核心数（比如4核显示是400%）', date_trunc, recent, 'sum(hosts.cpu_number)*100']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('cpu', 'XLPR服务器 CPU 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'cpu', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('cpu', body='<br/>'.join(htmltexts))
        return h

    def dbview_cpu_memory(self, recent=datetime.timedelta(days=1), date_trunc='hour'):
        from pyxllib.data.echarts import render_echart_html

        args = ['内存（单位：GB）', date_trunc, recent, 'sum(hosts.cpu_gb)']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('cpu_memory', 'XLPR服务器 内存 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'cpu_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('cpu_memory', body='<br/>'.join(htmltexts))
        return h

    def dbview_disk_memory(self, recent=datetime.timedelta(days=360), date_trunc='day'):
        """ 查看disk硬盘使用近况
        """
        from pyxllib.data.echarts import render_echart_html

        args = ['硬盘（单位：GB）', date_trunc, recent, 'sum(hosts.disk_gb)']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('disk_memory', 'XLPR服务器 DISK硬盘 使用近况', *args)
        htmltexts.append(res[0])
        htmltexts.append('注：xlpr4（四卡）服务器使用du计算/home大小有问题，未统计在列<br/>')

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'disk_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('disk_memory', body='<br/>'.join(htmltexts))
        return h

    def dbview_gpu_memory(self, recent=datetime.timedelta(days=30), date_trunc='day'):
        """ 查看gpu近使用近况

        TODO 这里有可以并行处理的地方，但这个方法不是很重要，不需要特地去做加速占用cpu资源
        """
        from pyxllib.data.echarts import render_echart_html

        args = ['显存（单位：GB）', date_trunc, recent, 'sum(hosts.gpu_gb)']

        htmltexts = [
            '<a target="_blank" href="https://www.yuque.com/xlpr/data/hnpb2g?singleDoc#"> 《服务器监控》工具使用文档 </a>']
        res = self._get_host_trace_total('gpu_memory', 'XLPR八台服务器 GPU显存 使用近况', *args)
        htmltexts.append(res[0])

        hosts = self.execute('SELECT host_name, nick_name FROM hosts WHERE gpu_gb > 0').fetchall()
        host_stats = []
        for i, (hn, nick_name) in enumerate(hosts, start=1):
            name = f'{hn}，{nick_name}' if nick_name else hn
            host_stats.append(self._get_host_trace_per_host(hn, 'gpu_memory', f'{name}', *args))
        host_stats.sort(key=lambda x: -x[1])  # 按使用量，从多到少排序
        htmltexts += [x[0] for x in host_stats]

        self.commit()

        h = render_echart_html('gpu_memory', body='<br/>'.join(htmltexts))
        return h

    def __4_一些数据更新操作(self):
        """ 比如一些扩展字段，在调用api的时候为了性能并没有进行计算，则可以这里补充更新 """

    def update_files_dhash(self, print_mode=True):
        """ 更新files表中的dhash字段值 """
        from pyxllib.cv.imhash import dhash
        from pyxllib.xlcv import xlpil

        # 获取总图片数
        total_count = self.execute("SELECT COUNT(*) FROM files WHERE dhash IS NULL").fetchall()[0][0]

        # 执行查询语句，获取dhash为NULL的记录

        # 初始化进度条
        progress_bar = tqdm(total=total_count, disable=not print_mode)

        while True:
            result = self.execute("SELECT id, data FROM files WHERE dhash IS NULL LIMIT 5")
            if not result:
                break
            for row in result:
                file_id = row[0]
                im = xlpil.read_from_buffer(row[1])
                computed_dhash = str(dhash(im))
                self.update_row('files', {'dhash': computed_dhash}, {'id': file_id})
                progress_bar.update(1)
            self.commit()

    def append_history(self, table_name, where, backup_keys, *,
                       can_merge=None,
                       update_time=None,
                       commit=False):
        """ 为表格添加历史记录，请确保这个表有一个jsonb格式的historys字段

        这里每次都会对关键字段进行全量备份，没有进行高级的优化。
        所以只适用于一些历史记录功能场景。更复杂的还是需要另外自己定制。

        :param table_name: 表名
        :param where: 要记录的id的规则，请确保筛选后记录是唯一的
        :param backup_keys: 需要备份的字段名
        :param can_merge: 在某些情况下，history不需要非常冗余地记录，可以给定与上一条合并的规则
            def can_merge(last, now):
                "last是上一条字典记录，now是当前要记录的字典数据，
                返回True，则用now替换last，并不新增记录"
                ...

        :param update_time: 更新时间，如果不指定则使用当前时间
        """
        # 1 获得历史记录
        ops = ' AND '.join([f'{k}=%s' for k in where.keys()])
        historys = self.exec2one(f'SELECT historys FROM {table_name} WHERE {ops}', list(where.values())) or []
        if historys:
            status1 = historys[-1]
        else:
            status1 = {}

        # 2 获得新记录
        if update_time is None:
            update_time = utc_timestamp()
        status2 = self.exec2dict(f'SELECT {",".join(backup_keys)} FROM {table_name} WHERE {ops}',
                                 list(where.values())).fetchone()
        status2['update_time'] = update_time

        # 3 添加历史记录
        if can_merge is None:
            def can_merge(status1, status2):
                for k in backup_keys:
                    if status1.get(k) != status2.get(k):
                        return False
                return True

        if historys and can_merge(status1, status2):
            historys[-1] = status2
        else:
            historys.append(status2)

        self.update_row(table_name, {'historys': historys}, where, commit=commit)



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import time

__dir__ = os.path.dirname(os.path.abspath(__file__))

sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '..')))

from pyxlpr.ppocr.data import build_dataloader
from pyxlpr.ppocr.modeling.architectures import build_model
from pyxlpr.ppocr.postprocess import build_post_process
from pyxlpr.ppocr.metrics import build_metric
from pyxlpr.ppocr.utils.save_load import load_model
from pyxlpr.ppocr.utils.utility import print_dict
import pyxlpr.ppocr.tools.program as program


def main(config, device, logger):
    global_config = config['Global']
    # build dataloader
    valid_dataloader = build_dataloader(config, 'Eval', device, logger)

    # build post process
    post_process_class = build_post_process(config['PostProcess'],
                                            global_config)

    # build model
    # for rec algorithm
    if hasattr(post_process_class, 'character'):
        char_num = len(getattr(post_process_class, 'character'))
        if config['Architecture']["algorithm"] in ["Distillation",
                                                   ]:  # distillation model
            for key in config['Architecture']["Models"]:
                config['Architecture']["Models"][key]["Head"][
                    'out_channels'] = char_num
        else:  # base rec model
            config['Architecture']["Head"]['out_channels'] = char_num

    model = build_model(config['Architecture'])
    extra_input = config['Architecture'][
        'algorithm'] in ["SRN", "NRTR", "SAR", "SEED"]
    if "model_type" in config['Architecture'].keys():
        model_type = config['Architecture']['model_type']
    else:
        model_type = None

    best_model_dict = load_model(config, model)
    if len(best_model_dict):
        logger.info('metric in ckpt ***************')
        for k, v in best_model_dict.items():
            logger.info('{}:{}'.format(k, v))

    # build metric
    eval_class = build_metric(config['Metric'])
    # start eval
    metric = program.eval(model, valid_dataloader, post_process_class,
                          eval_class, model_type, extra_input)
    logger.info(str(metric))
    return metric


if __name__ == '__main__':
    config, device, logger, vdl_writer = program.preprocess()
    main(config, device, logger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/03/28

"""
处理数据文件常用的拆分逻辑

半定制化的功能组件
"""

from collections import defaultdict, Counter
import datetime
import re
import json

from tqdm import tqdm
from joblib import Parallel, delayed

from pyxllib.prog.pupil import check_counter, tprint, typename
from pyxllib.file.specialist import XlPath, refinepath


def get_timestamp(fmt='%Y%m%d%H%M%S'):
    return datetime.datetime.now().strftime(fmt)


def process_file_base(file_func, file, correct_tag, error_tag,
                      *,
                      correct_dir=None, error_dir=None,
                      correct_file_name=None, error_file_name=None,
                      split_group_func=None,
                      reset=False):
    """ 处理单个文件数据，默认将两套运行结果放在同目录的两个文件中

    :param file_func: 对于自定义的处理函数
        返回支持1个参数，2个参数。1个参数时，默认没有错误的记录数据。
        对于数据，支持list[str]或list[dict]等多种格式
    :param split_group_func: 对正确的数据，按照某种规则进一步分组
    """

    # 0 工具函数
    def remove_old_file(new_file):
        # 如果目录下有同名函数，仅最后的时间戳不同的旧文件，删除
        stem = re.sub(r'_\d{14}$', '', new_file.stem)
        fmt_str = fr'{stem}_\d{{14}}{new_file.suffix}'
        for f in new_file.parent.glob(f'{stem}_*{new_file.suffix}'):
            if re.match(fmt_str, f.name):
                f.delete()

    def write_file(_dir, name, tag, data):
        if name is None:
            suffix = file.suffix if isinstance(data[0], str) else '.jsonl'
            name = f'{file.stem}_{tag}{len(data)}_{get_timestamp()}{suffix}'
        _file = _dir / name
        _file.parent.mkdir(exist_ok=True, parents=True)
        if reset:  # 找到旧后缀的文件，删除，排除时间戳差异的文件检索
            remove_old_file(_file)

        if isinstance(data[0], str):
            _file.write_text('\n'.join(data))
        else:
            _file.write_jsonl(data)

    # 1 处理数据
    file = XlPath(file)
    res = file_func(file)
    if isinstance(res, tuple) and len(res) == 2:
        correct_data, error_data = res
    else:
        correct_data = res
        error_data = None

    # 2 计算标准文件名
    if correct_data:
        correct_dir = file.parent if correct_dir is None else XlPath(correct_dir)
        suffix = file.suffix if isinstance(correct_data[0], str) else '.jsonl'
        if split_group_func:  # 使用split_group_func的时候，split_group_func不生效
            group_data = defaultdict(list)
            for x in correct_data:
                group_data[split_group_func(x)].append(x)

            for k, correct_data in group_data.items():
                correct_file_name2 = (f'{file.stem}_{refinepath(k)}_'
                                      f'{correct_tag}{len(correct_data)}_{get_timestamp()}{suffix}')
                write_file(correct_dir, correct_file_name2, correct_tag, correct_data)
        else:
            write_file(correct_dir, correct_file_name, correct_tag, correct_data)
    if error_data:
        error_dir = file.parent if error_dir is None else XlPath(error_dir)
        write_file(error_dir, error_file_name, error_tag, error_data)


def process_dir_base(file_func, dir_path, correct_tag, error_tag,
                     *, pattern='*', correct_dir=None, error_dir=None, reset=False,
                     **kwargs):
    """ 处理一个目录下的所有文件
    """
    dir_path = XlPath(dir_path)
    correct_dir = XlPath(correct_dir) if correct_dir else (dir_path.parent / f'{dir_path.name}_{correct_tag}')
    error_dir = XlPath(error_dir) if error_dir else (dir_path.parent / f'{dir_path.name}_{error_tag}')

    files = list(dir_path.rglob(pattern))
    for idx, file in enumerate(files, start=1):
        tprint(f'处理完第{idx}/{len(files)}个文件: {file.name} ==> {correct_tag}')
        process_file_base(file_func, file, correct_tag, error_tag,
                          correct_dir=correct_dir, error_dir=error_dir, reset=reset,
                          **kwargs)


def process_path(file_func, tag, path, **kwargs):
    """ 对单文件，或者目录处理的封装 """
    path = XlPath(path)

    if isinstance(file_func, str):  # 用命令行等接口的时候，输入可能是字符串名
        file_func = globals()[file_func]

    # 有些特殊参数，是预设给func使用的
    func_args = {}
    if 'remove_repeat_mode' in kwargs:
        func_args['remove_repeat_mode'] = kwargs.pop('remove_repeat_mode')

    if func_args:
        file_func2 = lambda x: file_func(x, **func_args)
    else:
        file_func2 = file_func

    if path.is_file():
        process_file_base(file_func2, path, tag, f'{tag}error', **kwargs)
    elif path.is_dir():
        process_dir_base(file_func2, path, tag, f'{tag}error', **kwargs)


class Analyzer:
    """ 分析器，一般用来写一些不影响主体功能，协助调试，分析数据特征的功能 """

    def __init__(self):
        self.counters = defaultdict(Counter)

    def count_once(self, tag, key):
        # tag相当于需要分组统计的名称
        self.counters[tag][key] += 1

    def check_counter(self):
        # 检查各种特征出现的次数
        for name, ct in self.counters.items():
            print(f'【{name}】')
            check_counter(ct)


def head_data(infile, num=1000, file_size=50):
    """ 获取infile头部部分数据

    :param file_size: 除了数量限制外，同时限制文件大小不超过file_size MB
        这里不是精确解法，而是简化大概的写法
    """
    res = []
    total_size = 0
    for x in XlPath(infile).yield_line(end=num):
        sz = len(x)
        total_size += sz
        if total_size > file_size * 1024 * 1024:
            break
        res.append(x)

    return res


def remove_repeat_base(infile,
                       get_etags_func,
                       exists_etags=set(),  # trick: 这里需要全局记录已出现过的etag
                       ):
    """ 去除重复数据，基础函数

    :param get_etags_func: 每道题目的etag，用于判断是否重复
        可能用于判断的etags数量不止一个，可以返回多个
        返回一个值的版本也可以，会自动转换为list
    """
    data = XlPath(infile).read_jsonl()
    src_len = len(data)

    data2 = []
    for x in data:
        etags = get_etags_func(x)
        if isinstance(etags, str):
            etags = [etags]
        if any(etag in exists_etags for etag in etags):
            continue
        data2.append(x)
        exists_etags.update(set(etags))

    new_len = len(data2)
    print(f'去重后剩余 {new_len}/{src_len} ≈ {new_len / src_len:.2%} 的数据')
    return data2


class CacheJsonlFile:
    """ 流式存储文件 """

    def __init__(self, parent_dir, prefix_stem, tag, batch_size=2000):
        self.parent_dir = parent_dir
        self.prefix_stem = prefix_stem
        self.tag = tag
        self.cache_text_lines = []
        self.batch_size = batch_size

        # 缓存时的文件名
        self.cache_file = XlPath(parent_dir) / f'{prefix_stem}_{tag}.cache.jsonl'
        self.total = 0

    def append(self, data):
        for x in data:
            if isinstance(x, str):
                self.cache_text_lines.append(x)
            else:
                self.cache_text_lines.append(json.dumps(x, ensure_ascii=False))
        if len(self.cache_text_lines) >= self.batch_size:
            self.flush()

    def flush(self):
        """ 刷新，将当前缓存写入文件 """
        if self.cache_text_lines:
            if self.total == 0:  # 第一次写入时，删除旧缓存文件
                self.cache_file.delete()

            self.total += len(self.cache_text_lines)
            self.parent_dir.mkdir(exist_ok=True, parents=True)
            with open(self.cache_file, 'a', encoding='utf8') as f:
                f.write('\n'.join(self.cache_text_lines) + '\n')
            self.cache_text_lines = []

    def save_all(self):
        """ 最终存储的文件名 """
        self.flush()
        if self.cache_file.is_file():
            dst_file = self.cache_file.with_stem(f'{self.prefix_stem}_{self.tag}{self.total}_{get_timestamp()}')
            self.cache_file.rename(dst_file)


class CacheJsonlGroupFiles:
    def __init__(self, parent_dir, prefix_stem, tag, group_func=None, batch_size=2000):
        self.files = {}

        self.parent_dir = parent_dir
        self.prefix_stem = prefix_stem
        self.tag = tag
        self.group_func = group_func
        self.batch_size = batch_size

    def append(self, data):
        groups = defaultdict(list)
        if self.group_func:
            for x in data:
                groups[self.group_func(x)].append(x)
        else:
            groups[''] = data

        for k, v in groups.items():
            if k not in self.files:
                subtag = f'{k}_{self.tag}' if k not in ('', None) else self.tag
                self.files[k] = CacheJsonlFile(self.parent_dir, self.prefix_stem, subtag, self.batch_size)
            self.files[k].append(v)

    def save_all(self):
        for file in self.files.values():
            file.save_all()


def process_single_file(root,
                        infile,
                        tag,
                        row_func,
                        *,
                        cur_idx=1,
                        total=None,
                        if_exists='skip',
                        group_func=None,
                        batch_size=2000,
                        debug=False,
                        show_tqdm=False):
    def finished():
        if total:
            tprint(f'处理第{cur_idx}/{total}个文件：{infile.name} ==> {tag}')
        else:
            tprint(f'处理文件：{infile.name} ==> {tag}')

    # 1 缓存路径
    srcdir = XlPath(root)
    infile = XlPath.init(infile, root=srcdir)
    relpath = infile.relative_to(srcdir)

    dstdir = srcdir.parent / f'{srcdir.name}_{tag}' / relpath.parent
    errdir = srcdir.parent / f'{srcdir.name}_{tag}error' / relpath.parent

    # 2 判断是不是已处理过的文件
    # stem = re.split(r'\d+_\d{14}', relpath.stem)[0]
    stem = relpath.stem + f'_{tag}'
    dstfiles = list(dstdir.glob_files(f'{stem}*.jsonl'))
    errfiles = list(errdir.glob_files(f'{stem}*.jsonl'))

    # 需要进一步过滤
    def is_exists_old_file(f):
        if f.name.endswith('.cache.jsonl'):
            return False
        stem2 = f.name[len(stem):]
        if re.match(r'(error)?\d+_\d{14}.jsonl$', stem2):
            return True

    # cache文件不算，不用管
    check_files = [f for f in (dstfiles + errfiles) if is_exists_old_file(f)]

    if check_files:
        if if_exists == 'skip':
            finished()
            return
        elif if_exists == 'overwrite':
            for f in check_files:
                f.delete()
        elif if_exists == 'error':
            raise FileExistsError(f'目标文件已存在：{check_files}')
        else:
            finished()
            return

    # 3 处理数据
    if debug:  # 如果开启调试模式，则关闭分组功能
        group_func = None
    dstcgf = CacheJsonlGroupFiles(dstdir, infile.stem, tag, group_func, batch_size)
    errcgf = CacheJsonlGroupFiles(errdir, infile.stem, tag + 'error', batch_size=batch_size)
    if debug:  # 如果开启调试模式，则单独分错误文件
        errcgf = dstcgf

    for line in tqdm(infile.yield_line(), desc=f'{infile.name} ==> {tag}', disable=not show_tqdm):
        # todo 出错的数据，应该添加错误信息，也存成一个新的jsonl格式
        row = row_func(line)
        if row and (not isinstance(row, dict) or row.get('status', 'ok') == 'ok'):
            # 必须要有内容，但如果内容有个status字段且不是'ok'，也不能要
            dstcgf.append([row])
        else:
            # 注意旧版存储的事line，但是新版的存储成row了
            # 但如果row确实没内容，还是兼容旧版，存储line
            errcgf.append([row or line])

    dstcgf.save_all()
    errcgf.save_all()
    finished()


class StructureAnalyzer:
    @classmethod
    def item_to_json(cls, x, depth):
        """ 获得字典结构的签名

        todo 这个项目之后，可以对这个函数进一步优化精简，作为以后解析结构的一个通用工具
        """
        if depth <= 0:
            return typename(x)

        if isinstance(x, dict):
            d = {}
            keys = sorted(x.keys())
            for k in keys:
                d[k] = cls.item_to_json(x[k], depth - 1)
        elif isinstance(x, list):
            d = []
            for k in x:
                d.append(cls.item_to_json(k, depth - 1))
        else:
            d = typename(x)

        return d

    @classmethod
    def item_to_str(cls, x, depth):
        res = cls.item_to_json(x, depth)
        return json.dumps(res)

    @classmethod
    def group_items(cls, items, depth):
        ct = Counter()
        groups = defaultdict(list)
        for x in items:
            desc = cls.item_to_str(x, depth)
            ct[desc] += 1
            groups[desc].append(x)
        # 按照值的数量对groups排序
        groups = dict(sorted(groups.items(), key=lambda x: -len(x[1])))
        return groups

    @classmethod
    def get_items_structures(cls, items, savefile=None):
        """ 获取jsonl数据的结构分布

        :param list[json] items: 一组json数据
        :return list[json]: 统计每种结构的分布数量，按树形结构，从多到少排序展示
        """

        def add_group(parent, items, depth, res):
            tag = parent['depth'] if parent else ''
            groups = cls.group_items(items, depth)
            for i, (desc, items2) in enumerate(groups.items(), start=1):
                if desc == parent['desc']:  # 再细化的结果跟父结点相同时，不再加深层级
                    continue

                d = {}
                d['depth'] = f'{tag}-{i}' if tag else f'{i}'
                d['count'] = len(items2)
                d['desc'] = desc
                d['structure'] = cls.item_to_json(items2[0], depth)
                res.append(d)

                add_group(d, items2, depth + 1, res)

        res = []  # 初始化结果列表
        add_group({'depth': '', 'desc': ''}, items, 1, res)  # 从空标签开始递归

        if savefile:
            XlPath(savefile).write_jsonl(res)

        return res

    @classmethod
    def compare_keys(cls, items):
        """ 对从 get_items_structures 获得的数据，对比structure的覆盖情况 """
        import pandas as pd
        from pyxllib.algo.stat import custom_fillna

        ls = []
        for x in items:
            row = {}
            row['__depth'] = x['depth']
            row['__count'] = x['count']
            if isinstance(x['structure'], dict):
                row.update(x['structure'])
            elif isinstance(x['structure'], list):
                row.update({f'[{k}]': v for k, v in enumerate(x['structure'])})
            ls.append(row)

        df = pd.DataFrame.from_dict(ls)
        # 筛选df，depth中没有'-'
        df = df[df['__depth'].apply(lambda x: '-' not in x)]
        df = custom_fillna(df).T

        return df


def process_batch_files(srcdir,
                        dsttag,
                        line_convert_func,
                        pattern='*.jsonl',
                        if_exists='skip',
                        processes_num=1,
                        group_func=None,
                        batch_size=2000,
                        debug=False,
                        ):
    """ 通用批处理函数

    :param srcdir: 输入待处理的目录或文件
    :param pattern: 文件名检索逻辑，如 '*'，'*.jsonl', '*_std*.jsonl'等
    :param if_exists: 如果目标文件已存在，如何处理。'skip'跳过不重复处理，'overwrite'覆盖重新运行，'error'抛出报错
    :param processes_num: 并发处理的进程数
    :param batch_size: 每个文件缓存的条目数，超过这个数目后就会先写入文件
    """
    # 1 检索文件
    srcdir = XlPath(srcdir)
    if srcdir.is_file():
        files = [srcdir.name]
        srcdir = srcdir.parent
    else:
        files = [f.relpath(srcdir) for f in srcdir.rglob_files(pattern)]

    file_num = len(files)

    # 2 并发处理多个文件
    backend = 'loky' if processes_num > 1 else 'sequential'
    tasks = []
    for i, f in enumerate(files, start=1):
        task = delayed(process_single_file)(srcdir,
                                            f,
                                            dsttag,
                                            line_convert_func,
                                            cur_idx=i,
                                            total=file_num,
                                            if_exists=if_exists,
                                            group_func=group_func,
                                            batch_size=batch_size,
                                            debug=debug,
                                            show_tqdm=processes_num == 1)
        tasks.append(task)

    Parallel(n_jobs=processes_num, backend=backend)(tasks)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/WenmuZhou/DBNet.pytorch/blob/master/data_loader/modules/iaa_augment.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import imgaug
import imgaug.augmenters as iaa


class AugmenterBuilder(object):
    def __init__(self):
        pass

    def build(self, args, root=True):
        if args is None or len(args) == 0:
            return None
        elif isinstance(args, list):
            if root:
                sequence = [self.build(value, root=False) for value in args]
                return iaa.Sequential(sequence)
            else:
                return getattr(iaa, args[0])(
                    *[self.to_tuple_if_list(a) for a in args[1:]])
        elif isinstance(args, dict):
            cls = getattr(iaa, args['type'])
            return cls(**{
                k: self.to_tuple_if_list(v)
                for k, v in args['args'].items()
            })
        else:
            raise RuntimeError('unknown augmenter arg: ' + str(args))

    def to_tuple_if_list(self, obj):
        if isinstance(obj, list):
            return tuple(obj)
        return obj


class IaaAugment():
    """ iaa名称取自：import imgaug.augmenters as iaa
    这里有些图像增强方法，可以通过配置文件指定

    注意这个增广，标注也会跟着变

    底层是用imgaug库实现的，具体原理结构还没深究
    """

    def __init__(self, augmenter_args=None, **kwargs):
        if augmenter_args is None:
            augmenter_args = [{
                'type': 'Fliplr',
                'args': {
                    'p': 0.5
                }
            }, {
                'type': 'Affine',
                'args': {
                    'rotate': [-10, 10]
                }
            }, {
                'type': 'Resize',
                'args': {
                    'size': [0.5, 3]
                }
            }]
        self.augmenter = AugmenterBuilder().build(augmenter_args)

    def __call__(self, data):
        image = data['image']
        shape = image.shape

        if self.augmenter:
            aug = self.augmenter.to_deterministic()
            data['image'] = aug.augment_image(image)
            data = self.may_augment_annotation(aug, data, shape)
        return data

    def may_augment_annotation(self, aug, data, shape):
        if aug is None:
            return data

        line_polys = []
        for poly in data['polys']:
            new_poly = self.may_augment_poly(aug, shape, poly)
            line_polys.append(new_poly)
        data['polys'] = np.array(line_polys)
        return data

    def may_augment_poly(self, aug, img_shape, poly):
        keypoints = [imgaug.Keypoint(p[0], p[1]) for p in poly]
        keypoints = aug.augment_keypoints(
            [imgaug.KeypointsOnImage(
                keypoints, shape=img_shape)])[0].keypoints
        poly = [(p.x, p.y) for p in keypoints]
        return poly



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/09/26 09:47

""" 单位功能库 """

import pint

import pandas as pd

from pyxllib.prog.newbie import xl_format_g
from pyxllib.prog.pupil import run_once


@run_once
def get_ureg():
    """ 如果不想重复生成，可以用这个函数，只定义一个 """
    ureg = pint.UnitRegistry()
    return ureg


def compare_quantity(units, base_unit, scale=None):
    """ 同一物理维度的数值量级大小对比

    比如一套天文长度、距离数据，对比它们之间的量级大小

    :param List[str, str] units: 一组要对比的单位数值
        第1个str是描述，第2个str是数值和单位
    :param base_unit: 基础单位值，比如长度是m，也可以用km等自己喜欢的基础单位值
    :param scale: 对数值进行缩放
    """
    ureg = get_ureg()

    # 统一到单位：米
    for x in units:
        x[1] = ureg(x[1]).to(base_unit)

    # 从小到大排
    units.sort(key=lambda x: x[1])

    # 格式化输出r
    unit_list = []
    if scale:
        columns = ['name', f'基础单位：{base_unit}', 'scale', '与上一项倍率']
    else:
        columns = ['name', f'基础单位：{base_unit}', 'scale', '与上一项倍率']
    last_value = units[0][1]
    for i, x in enumerate(units, start=1):
        x = x.copy()
        # x.append('{:.3g}'.format(round(float(x[1] / last_value), 2)))
        if scale:
            x.append(f'{x[1] * scale:.2e}')
        x.append(xl_format_g(round(float(x[1] / last_value), 2)))
        last_value = x[1]
        x[1] = f'{x[1]:.2e}'
        # print(f'{i:>4}', '\t'.join(x))
        unit_list.append(x)

    df = pd.DataFrame.from_records(unit_list, columns=columns)
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2024/10/12

"""
我们自己的一套利用后端服务进行文件同步的工具
"""

import os
import time

from loguru import logger
from tqdm import tqdm
import requests

from pyxllib.file.specialist import XlPath, GetEtag
from pyxllib.prog.pupil import format_exception


class SyncFileClient:
    """

    注意：该类填写的路径，一律是相对local_root、remote_root的路径，不是当前实际工作目录下的路径！
        可以通过把local_host设为空的来实现一些特殊精细化的操作
        但一般推荐还是都把local_root、remote_root设置上
    """

    def __1_basic(self):
        pass

    def __init__(self, host, token=None):
        """
        :param host: 可以只写主机名
        """
        token = token or os.getenv("XL_COMMON_SERVER_TOKEN")
        self.headers = {
            'Authorization': f'Bearer {token}',
        }
        self.local_root = None
        self.remote_root = None
        self.host = self.link_host(host)

    def link_host(self, hostname):
        from requests.exceptions import Timeout

        host = f'{os.getenv("MAIN_WEBSITE")}/{hostname}'
        resp = requests.get(f'{host}/common/get_local_server', headers=self.headers)
        ip = resp.json()['ip']

        if ip:
            try:
                resp = requests.get(f'{ip}/healthz', timeout=5)
                if resp.status_code == 200 and resp.json()['status'] == 'ok':
                    host = ip
            except Timeout:
                pass
            except Exception as e:
                raise e

        remote_root = requests.get(f'{host}/common/get_wkdir', headers=self.headers).json()['wkdir']
        self.remote_root = XlPath(remote_root)

        return host

    def get_abs_local_path(self, local_path, remote_path=None):
        # 1 如果需要，从远程路径推测本地路径
        if local_path is None and remote_path is not None:
            local_path = remote_path

        # 2 本地存储的文件路径
        if self.local_root:
            return self.local_root / local_path
        else:
            return XlPath(local_path)

    def get_abs_remote_path(self, remote_path, local_path=None):
        # 1 如果需要，从本地路径推测远程路径
        if remote_path is None and local_path is not None:
            remote_path = local_path

        # 2 远程存储的文件路径
        if self.remote_root:
            return self.remote_root / remote_path
        else:
            return XlPath(remote_path)

    def __2_sync(self):
        pass

    def upload_file(self, local_file=None, remote_file=None, relpath=True):
        """
        :param str remote_file: 要存储的远程文件路径，默认可以不指定，自动通过local_file存储到对称位置

        使用示例：
        sfc = SyncFileClient('codepc_mi15')
        sfc.upload_file('abc/README.md')
        sfc.upload_file('abc/README.md', 'bcd/1.md')
        """
        file = [local_file]
        if remote_file:
            file.append(remote_file)
        return self.upload_files([file], relpath=relpath)

    def upload_files(self, files, relpath=True):
        """
        :param list[local_file<, remote_path>] files:
            remote_path可选，未写的时候会自动填充
        :return:

        todo 加进度条？不过问了下gpt，这个实现好像有点复杂，先不折腾了。本来其实就推荐单个单个文件上传，在外部进行进度展示管控。

        使用示例：
        sfc = SyncFileClient('codepc_mi15')
        print(sfc.upload_files(['data/d231120禅宗1期4阶.xlsx', 'data/d231120禅宗1期4阶2.xlsx']))
        """
        # 1 映射出目标路径
        files2 = []
        for file in files:
            if not isinstance(file, (list, tuple)):
                local_file, remote_file = file, None
            else:
                local_file, remote_file = file

            if relpath:
                local_file = self.get_abs_local_path(local_file, remote_file)
                remote_file = self.get_abs_remote_path(remote_file, local_file)

            files2.append([local_file, remote_file.as_posix()])

        files3 = [('files', (remote_file, open(local_file, 'rb'))) for local_file, remote_file in files2]

        # 4 发送请求
        resp = requests.post(f'{self.host}/common/upload_files',
                             headers=self.headers,
                             files=files3)
        return resp.json()

    def download_file(self, remote_file=None, local_file=None, relpath=True):
        """
        :param str remote_file: 要下载的远程文件路径，相对get_wkdir下的路径
        :param str local_file: 下载到本地的文件路径，默认可以不指定，自动通过remote_file存储到对称位置

        todo 加进度条？方便超大文件下载查看进度。

        使用示例：
        sfc = SyncFileClient('http://yourserver.com', headers={"Authorization": "Bearer your_token"})
        sfc.download_file('.vscode/launch.json', 'launch.json')
        """
        if relpath:
            local_file = self.get_abs_local_path(local_file, remote_file)
            remote_file = self.get_abs_remote_path(remote_file, local_file)

        data = {'file': remote_file.as_posix()}
        # 使用 stream=True 开启流式处理
        with requests.post(f'{self.host}/common/download_file',
                           headers=self.headers, json=data, stream=True) as resp:
            if resp.status_code == 200:
                # 以二进制方式写入文件
                with open(local_file, 'wb') as f:
                    # 分块写入文件，每次 1MB (1024 * 1024 bytes)
                    for chunk in resp.iter_content(chunk_size=1024 * 1024):
                        if chunk:  # 忽略keep-alive的空块
                            f.write(chunk)
                return local_file
            else:
                # 处理其他状态码，抛出异常或进行错误处理
                print(f"Error: Unable to download file, status code {resp.status_code}")
                return None

    def verify_etag(self, local_file=None, remote_file=None, relpath=True):
        """
        :param relpath: 是否使用的是相对路径
            当不使用relpath的时候，注意local_file、remote_file都是手动必填的XlPath类型字段
        :return:
        """
        if relpath:
            local_file = self.get_abs_local_path(local_file, remote_file)
            remote_file = self.get_abs_remote_path(remote_file, local_file)
        if not local_file.is_file():
            return {'match': False}
        local_etag = GetEtag.from_file(local_file)
        json_data = {'file': remote_file.as_posix(), 'etag': local_etag}
        resp = requests.post(f'{self.host}/common/check_file_etag',
                             headers=self.headers,
                             json=json_data)
        return resp.json()

    def upload_dir(self,
                   local_dir=None,
                   remote_dir=None,
                   *,
                   delete_local_file=False,
                   verify_etag=False,
                   relpath=True):
        """
        :param local_dir: 本地目录
        :param remote_dir: 远程目录
        :param delete_local_file: 每个文件上传成功后，是否删除本地文件
            默认不删除，一般不删除是配合etag校验使用的
        :param verify_etag: 上传前是否校验etag，如果服务器上对应位置已经有对应etag的文件，其实就不用上传了

        该方法已修改以支持上传子目录和空目录
        """
        if relpath:
            # D:\home\chenkunze\data\temp2，举个具体的例子案例值，供参考理解
            local_dir = self.get_abs_local_path(local_dir, remote_dir)
            # /home/chenkunze/data/temp2
            remote_dir = self.get_abs_remote_path(remote_dir, local_dir)

        for root, dirs, files in os.walk(local_dir):
            # root拿到的是这样的绝对路径：D:\home\chenkunze\data\temp2\数值问题
            # '数值问题'
            subdir = XlPath(root).relpath(local_dir)
            if not files and not dirs:
                # 如果该目录为空，则在远程服务器上创建空目录
                self.create_remote_dir(remote_dir / subdir, relpath=False)

            for file in tqdm(files, desc=f'上传目录：{root}'):
                local_path = local_dir / subdir / file
                remote_path = remote_dir / subdir / file
                if verify_etag and self.verify_etag(local_path, remote_path, relpath=False):
                    pass  # 如果etag校验通过，跳过上传
                else:
                    self.upload_file(local_path, remote_path, relpath=False)

                if delete_local_file:
                    os.remove(local_path)

    def create_remote_dir(self, remote_dir, relpath=True):
        """
        创建远程目录的辅助方法
        """
        if relpath:
            remote_dir = self.get_abs_remote_path(remote_dir)
        json_data = {'dir': remote_dir.as_posix()}
        resp = requests.post(f'{self.host}/common/create_dir',
                             headers=self.headers,
                             json=json_data)
        return resp.json()

    def download_dir(self, remote_dir=None, local_dir=None, *, verify_etag=False, relpath=True):
        """
        下载远程目录到本地
        :param remote_dir: 远程目录路径，相对路径或绝对路径都支持
        :param local_dir: 本地保存目录，默认会自动映射到对应的远程位置
        :param verify_etag: 是否进行etag校验，跳过已经在本地存在并且未修改的文件
        :return:

        示例：
        sfc = SyncFileClient('codepc_mi15')
        sfc.download_dir('data/temp2')
        """
        # 1. 获取绝对路径
        if relpath:
            local_dir = self.get_abs_local_path(local_dir, remote_dir)
            remote_dir = self.get_abs_remote_path(remote_dir, local_dir)

        # 2. 从服务器获取远程目录的文件和子目录结构
        json_data = {'dir': remote_dir.as_posix()}
        resp = requests.post(f'{self.host}/common/list_dir', headers=self.headers, json=json_data)
        if resp.status_code != 200:
            raise Exception(f"Failed to list remote directory: {resp.text}")

        dir_structure = resp.json()  # 期望返回格式为 {'dirs': [...], 'files': [...]}

        # 3. 创建本地目录
        if not local_dir.exists():
            local_dir.mkdir(parents=True)

        # 4. 先处理子目录
        for subdir in dir_structure.get('dirs', []):
            local_subdir = local_dir / subdir
            if not local_subdir.exists():
                local_subdir.mkdir(parents=True)

            # 递归下载子目录中的文件
            for i in range(100):  # 开发机senseserver3好像不稳定，加了重试机制
                try:
                    self.download_dir(remote_dir / subdir, local_dir / subdir, verify_etag=verify_etag)
                    break
                except requests.exceptions.RequestException as e:
                    logger.warning(format_exception(e, 3))
                    time.sleep(60)
                    continue
            else:
                raise requests.exceptions.RequestException

        # 5. 下载文件
        for file in tqdm(dir_structure.get('files', []), desc=f'下载目录：{remote_dir}'):
            remote_file = remote_dir / file
            local_file = local_dir / file

            for i in range(100):
                try:
                    # 校验etag，如果verify_etag为True并且校验成功，则跳过下载
                    if verify_etag and self.verify_etag(local_file, remote_file, relpath=False)['match']:
                        break

                    # 下载文件
                    logger.info(f'下载文件中：{local_file}')
                    self.download_file(remote_file, local_file, relpath=False)
                    break
                except requests.exceptions.RequestException:
                    time.sleep(60)
                    continue
            else:
                raise requests.exceptions.RequestException

    def download_path(self, remote_path=None, local_path=None, *, verify_etag=False, relpath=True):
        """
        下载远程文件或目录到本地
        :param remote_path: 远程路径，可以是文件或者目录
        :param local_path: 本地保存路径，默认会自动映射到对应的远程位置
        :param verify_etag: 是否进行etag校验，跳过已经在本地存在并且未修改的文件
        :return:
        """
        if relpath:
            # 获取远程路径的绝对路径
            remote_path = self.get_abs_remote_path(remote_path, local_path)
            local_path = self.get_abs_local_path(local_path, remote_path)

        # 判断是文件还是目录
        json_data = {'path': remote_path.as_posix()}
        resp = requests.post(f'{self.host}/common/check_path_type', headers=self.headers, json=json_data)

        if resp.status_code != 200:
            raise Exception(f"Failed to check remote path type: {resp.text}")

        path_info = resp.json()

        if path_info['type'] == 'file':
            # 如果是文件，调用 download_file
            self.download_file(remote_file=remote_path, local_file=local_path, relpath=False)
        elif path_info['type'] == 'dir':
            # 如果是目录，调用 download_dir
            self.download_dir(remote_dir=remote_path, local_dir=local_path,
                              verify_etag=verify_etag, relpath=False)
        else:
            raise ValueError(f"Unknown path type: {path_info['type']}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2022/04/07 15:45

"""
树形结构相关的处理
"""

import re

from anytree import Node


class XlNode(Node):

    def sign_node(self, check_func, *, flag_name='_flag', child_depth=-1, reset_flag=False):
        """ 遍历所有结点

        :param check_func: 检查node是否符合要求
            True，自身打上标记2，其余父节点、子结点全部打上1 （这里处理方式后续可以再扩展更灵活）

        不是那么通用，但也可能存在复用的功能，做了个封装
        """
        from anytree import PreOrderIter

        if child_depth == -1:
            child_depth = None
        else:
            child_depth += 1

        if reset_flag:
            for x in PreOrderIter(self):
                setattr(x, flag_name, 0)

        def set_flag(node, flag):
            # 结点没有值或者flag小余当前值时，才标记更新为当前flag
            setattr(node, flag_name, max(flag, getattr(node, flag_name, 0)))

        cnt = 0
        for x in PreOrderIter(self, filter_=check_func):
            # 对找到的结点标记2
            cnt += 1
            set_flag(x, 2)
            # 其父结点、子结点标记1
            for y in x.ancestors:
                set_flag(y, 1)
            for y in PreOrderIter(x, maxlevel=child_depth):
                set_flag(y, 1)
        return cnt

    def render(self, attrname='name', *,
               childiter=list, maxlevel=None, filter_=None,  # 结点的迭代、过滤等规则
               prefix='\t',
               dedent=0):
        """ 渲染展示结构树

        :pram attrname:
            str, 要展示的成员变量值
            def func(node, depth)，自定义展示样式
        :param prefix: 缩进样式
            str，每个层级增加的样式，会直接用 prefix*depth作为前缀
            tuple(3), 长度3的tuple，对应AbstractStyle(vertical, cont, end)
            def func(depth), 输入深度，自定义前缀样式 （未实装，想到这个理念，但觉得用处不大）
        :param dedent: 统一减少缩进层级

        """
        from anytree import RenderTree

        # 1 文本格式化的规则
        if isinstance(attrname, str):
            def to_str(node, depth):
                if depth < 0:
                    return ''
                return getattr(node, attrname)
        elif callable(attrname):
            to_str = attrname
        else:
            raise ValueError

        # 2 遍历结点
        ls = []
        for pre, fill, node in RenderTree(self, childiter=childiter, maxlevel=maxlevel):
            if filter_ and not filter_(node):
                continue

            depth = len(pre) // 4 - dedent
            ls.append(prefix * depth + to_str(node, depth))

        return '\n'.join(ls)

    def render_html(self, attrname='name', **kwargs):
        """ 渲染成html页面内容
        """
        if isinstance(attrname, str):
            def attrname(node, depth):
                if depth < 0:
                    return '<br/>'
                content = getattr(node, attrname)
                div = f'<div style="padding-left:{depth * 2 + 1}em;text-indent:-1em">{content}</div>'
                return div

        return XlNode.render(self, attrname, **kwargs)

    def find_parent(self, check_func):
        """ 查找包含自身的所有父结点中，name符合check_func的结点 """
        if isinstance(check_func, str):
            def _check_func(x):
                return x.name == check_func
        elif isinstance(check_func, re.Pattern):
            def _check_func(x):
                return check_func.search(x.name)
        elif callable(check_func):
            _check_func = check_func
        else:
            raise ValueError

        p = self
        while p and not _check_func(p):
            p = p.parent
        return p

    def next_preorder_node(self, iter_child=True):
        """ 自己写的先序遍历

        主要应用在xml、bs4相关遍历检索时，有时候遇到特殊结点
            可能子结点不需要解析
            或者整个cur_node和子结点已经被解析完了，不需要再按照通常的先序遍历继续进入子结点
        此时可以 iter_child=False，进入下一个兄弟结点
        """
        from anytree.util import rightsibling

        if iter_child and self.children:
            return self.children[0]
        else:
            cur_node = self
            while True:
                parent = cur_node.parent
                if parent is None:
                    return None
                sibing = rightsibling(cur_node)
                if sibing:
                    return sibing
                cur_node = parent



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2023/11/05

""" 基于jieba库的一些文本处理功能 """

from collections import Counter
import re

from tqdm import tqdm
import pandas as pd

import jieba
import jieba.posseg as pseg
from simhash import Simhash

from pyxllib.prog.pupil import DictTool, run_once
from pyxllib.file.specialist import XlPath
from pyxllib.algo.stat import update_dataframes_to_excel


def jieba_add_words(words):
    for w in words:
        jieba.add_word(w)


def jieba_del_words(words):
    for w in words:
        jieba.del_word(w)


@run_once('str')
def jieba_cut(text):
    return tuple(jieba.cut(text))


@run_once('str')
def pseg_cut(text):
    return tuple(pseg.cut(text))


def _count_word_frequency(texts, function_word=True):
    """ 统计关键词出现频数 (主要是协助计算tf-idf)

    :param texts: 输入字符串列表
    :param function_word: 是否要统计虚词
    :return: 一个dict
        key: 分词名称
        values: [x, y]，x是出现总频数，y是这个词在多少篇文章中出现过

    >>> _count_word_frequency(['正正正正', '正反正', '反反反反'])
    {'正正': [1, 1], '反反': [2, 1]}

    原没有过滤词性的结果：{'正正': [2, 1], '正': [1, 1], '反正': [1, 1], '反反': [2, 1]}
    """

    d = dict()
    for text in tqdm(texts, '词频统计'):
        wordflags = list(pseg.cut(text))
        words = set()
        for word, flag in wordflags:
            # 虚词不做记录
            if (not function_word) and flag in ('uj', 'd', 'p', 'c', 'u', 'xc'):
                continue
            words.add(word)
            if word not in d:
                d[word] = [0, 0]
            d[word][0] += 1
        for word in words:
            d[word][1] += 1
    return d


def analyse_tf_idf(texts, outfile=None, sheet_name='tf-idf', *, function_word=True):
    """ 分析tf-idf值

    :param list[str] texts: 多份文件的文本内容
    :return: 一个DataFrame数据

    这个算法jieba可能有些自带库可以搞，但是自己写一下也不难啦
    注意我这里返回的tf-idf中，是放大了总频数倍的，这样显示的数值大一点，看起来舒服~
    """
    from math import log10

    frequency = _count_word_frequency(texts, function_word)
    DictTool.isub(frequency, [' ', '\t', '\n'])

    n = len(texts)
    sum_frequency = sum([v[0] for v in frequency.values()])

    li = []
    for k, v in frequency.items():
        idf = log10(n / v[1])
        # idf = 1
        li.append([k, v[0], v[0] / sum_frequency, v[1], idf, v[0] * idf])
    df = pd.DataFrame.from_records(li, columns=('词汇', '频数', '频率', '出现该词文章数', 'idf', 'tf-idf'))
    df.sort_values(by='tf-idf', ascending=False, inplace=True)

    if outfile:
        update_dataframes_to_excel(outfile, {sheet_name: df})

    return df


class TextClassifier:
    def __init__(self, texts=None):
        """ 文本分类器

        :param list[str] texts: 文本内容
        """

        self.texts = []
        self.tfidf = {}
        self.vecs = []  # 每份文本对应的向量化表达
        self.default_tfidf = 1  # 如果没有计算tf-idf，可以全部默认用权重1

        if texts:
            for text in texts:
                self.texts.append(text)

    def get_text_tf(self, text, *,
                    function_word_weight=0.2,
                    normalize=True,
                    ingore_words=(' ', '\t', '\n'),
                    add_flag=False):
        """ 这里可以定制提取text关键词的算法

        :param function_word_weight: 这里可以自定义功能性词汇权重，一般是设一个小数降低权重

        一般是定制一些过滤规则，比如过滤掉一些词性，或者过滤掉一些词
        """
        ct = Counter()

        # 1 初步的分词，以及是否要过滤虚词
        wordflags = list(pseg_cut(text))
        for word, flag in wordflags:
            if flag in ('uj', 'd', 'p', 'c', 'u', 'xc', 'x'):
                if add_flag:
                    ct[word + ',' + flag] += function_word_weight
                else:
                    ct[word] += function_word_weight
            else:
                if add_flag:
                    ct[word + ',' + flag] += 1
                else:
                    ct[word] += 1

        # 2 归一化一些词
        if normalize:
            ct2 = Counter()
            for k, v in ct.items():
                # 如果需要对一些词汇做归一化，也可以这里设置
                k = re.sub(r'\d', '0', k)  # 把数字都换成0
                ct2[k] += v
            ct = ct2

        # 3 过滤掉一些词
        if ingore_words:
            for k in ingore_words:
                if k in ct:
                    del ct[k]

        return ct

    def compute_tfidf(self, outfile=None, sheet_name='tf-idf', normalize=False, function_word_weight=0.2,
                      add_flag=False):
        """ 重算tfidf表 """
        from math import log10

        # 1 统计频数和出现该词的文章数
        d = dict()
        for text in tqdm(self.texts, '词频统计'):
            ct = self.get_text_tf(text, normalize=normalize, function_word_weight=function_word_weight,
                                  add_flag=add_flag)
            for k, v in ct.items():
                if k not in d:
                    d[k] = [0, 0]
                d[k] = [d[k][0] + v, d[k][1] + 1]

        # 2 计算tfidf
        n = len(self.texts)
        sum_tf = sum([v[0] for v in d.values()])
        ls = []
        for k, v in d.items():
            idf = log10(n / v[1])
            # idf = 1
            ls.append([k, v[0], v[0] / sum_tf, v[1], idf, v[0] * idf])

        df = pd.DataFrame.from_records(ls, columns=('词汇', '频数', '频率', '出现该词文章数', 'idf', 'tf-idf'))
        df.sort_values(by='tf-idf', ascending=False, inplace=True)

        # 3 保存到文件
        if outfile:
            update_dataframes_to_excel(outfile, {sheet_name: df})

        self.tfidf = {row['词汇']: row['tf-idf'] for idx, row in df.iterrows()}
        self.default_tfidf = df.loc[len(df) - 1]['tf-idf']  # 最后条的权重作为其他未见词的默认权重

        return df

    def normalization(self, d):
        """ 向量归一化

        输入一个类字典结构表示的向量，对向量做归一化处理
        """
        length = sum([v * v for v in d.values()]) ** 0.5  # 向量长度
        return {k: v / length for k, v in d.items()}

    def get_text_vec(self, text):
        """ 获取文本的向量化表达

        :param str text: 文本内容
        """
        ct = self.get_text_tf(text)
        vec = {k: v * self.tfidf.get(k, self.default_tfidf) for k, v in ct.items()}
        vec = self.normalization(vec)
        return vec

    def compute_vecs(self):
        """ 重置向量化表达 """
        vecs = []
        for text in tqdm(self.texts, desc='query向量化'):
            vecs.append(self.get_text_vec(text))
        self.vecs = vecs
        return vecs

    def cosine_similar(self, x, y):
        """ 两个向量的余弦相似度，值越大越相似

        这里是简化的，只算两个向量的点积，请确保输入的都是单位长度的向量
        注意这里x和y都是稀疏矩阵的存储形式，传入的是dict结构
        """
        keys = x.keys() & y.keys()  # 求出x和y共有的键值
        return sum([x[k] * y[k] for k in keys])

    def find_similar_vec(self, x, maxn=10):
        """ 找与x最相近的向量，返回下标和相似度

        :pamra x: 待查找的对象
        :param maxn: 返回最相近的前maxn个对象
        """
        if isinstance(x, str):
            x = self.get_text_vec(x)

        # todo 使用并行计算？或者其实也可以向量化，但向量化是稀疏矩阵，挺占空间的
        sims = [(i, self.cosine_similar(x, v)) for i, v in enumerate(self.vecs)]
        sims.sort(key=lambda x: x[1], reverse=True)
        return sims[:maxn]

    def refine_vecs(self):
        """ 优化向量数据，去掉权重小余0.0001的维度 """
        # 1 计算每个向量的长度
        vecs = []
        for vec in tqdm(self.vecs, '优化向量'):
            vec = [(k, v) for k, v in vec.items()]
            vec.sort(key=lambda x: x[1], reverse=True)
            vec2 = {}
            for k, v in vec:
                if v < 0.0001:
                    break
                vec2[k] = round(v, 4)
            vecs.append(vec2)

        self.vecs = vecs
        return self.vecs
